from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import func
from sqlalchemy import (
    select,
    Integer,
    ForeignKey,
    String,
    UniqueConstraint,
    Index,
    CheckConstraint,
    event,
    DDL,
    Column,
    TEXT,
    DOUBLE_PRECISION,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import relationship, column_property

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.optimization.record_db_models import (
    OptimizationSpecificationORM,
    OptimizationRecordORM,
    OptimizationTrajectoryORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import BaseORM

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class TorsiondriveOptimizationORM(BaseORM):
    """
    Table for storing torsiondrive to optimization relationships
    """

    __tablename__ = "torsiondrive_optimization"

    torsiondrive_id = Column(Integer, ForeignKey("torsiondrive_record.id", ondelete="cascade"), primary_key=True)
    optimization_id = Column(Integer, ForeignKey(OptimizationRecordORM.id), primary_key=True)
    key = Column(String, nullable=False, primary_key=True)
    position = Column(Integer, primary_key=True)

    energy = column_property(
        select(OptimizationRecordORM.energies[-1]).where(OptimizationRecordORM.id == optimization_id).scalar_subquery()
    )

    optimization_record = relationship(OptimizationRecordORM)

    __table_args__ = (Index("ix_torsiondrive_optimization_id", "optimization_id"),)

    _qcportal_model_excludes = ["torsiondrive_id"]


class TorsiondriveInitialMoleculeORM(BaseORM):
    """
    Table for storing torsiondrive to initial molecule relationships
    """

    __tablename__ = "torsiondrive_initial_molecule"

    torsiondrive_id = Column(Integer, ForeignKey("torsiondrive_record.id", ondelete="cascade"), primary_key=True)
    molecule_id = Column("molecule_id", Integer, ForeignKey(MoleculeORM.id), primary_key=True)

    molecule = relationship(MoleculeORM)

    _qcportal_model_excludes = ["torsiondrive_id"]


class TorsiondriveSpecificationORM(BaseORM):
    """
    Table for storing torsiondrive specifications
    """

    __tablename__ = "torsiondrive_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String(100), nullable=False)

    optimization_specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=False)
    optimization_specification = relationship(OptimizationSpecificationORM, lazy="joined")

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "specification_hash", "optimization_specification_id", name="ux_torsiondrive_specification_keys"
        ),
        Index("ix_torsiondrive_specification_program", "program"),
        Index("ix_torsiondrive_specification_optimization_specification_id", "optimization_specification_id"),
        # Enforce lowercase on some fields
        # This does not actually change the text to lowercase, but will fail to insert anything not lowercase
        # WARNING - these are not autodetected by alembic
        CheckConstraint("program = LOWER(program)", name="ck_torsiondrive_specification_program_lower"),
    )

    # TODO - protocols will be in the model eventually
    _qcportal_model_excludes = [
        "id",
        "specification_hash",
        "optimization_specification_id",
        "protocols",
    ]

    @property
    def short_description(self) -> str:
        return f"{self.program}~{self.optimization_specification.short_description}"


# CTE for a table with minimimum optimizations. Has columns torsiondrive_id, key, and (minimum) optimization_id
# Chooses the optimization with the lowest energy, and if there are multiple, the one with the lowest id
_minopt_cte = (
    select(
        TorsiondriveOptimizationORM.torsiondrive_id.label("torsiondrive_id"),
        TorsiondriveOptimizationORM.key.label("key"),
        TorsiondriveOptimizationORM.optimization_id.label("optimization_id"),
    )
    .join(OptimizationRecordORM, TorsiondriveOptimizationORM.optimization_id == OptimizationRecordORM.id)
    .distinct(TorsiondriveOptimizationORM.torsiondrive_id, TorsiondriveOptimizationORM.key)
    .order_by(
        TorsiondriveOptimizationORM.torsiondrive_id,
        TorsiondriveOptimizationORM.key,
        OptimizationRecordORM.energies[-1].cast(TEXT).cast(DOUBLE_PRECISION).asc(),
        OptimizationRecordORM.id.asc(),
    )
    .cte()
)

# CTE for a table with minimimum optimizations, but as JSON. Has columns torsiondrive_id, minimum_optimizations (as JSONB)
_minopt_cte_agg = (
    select(
        _minopt_cte.c.torsiondrive_id.label("torsiondrive_id"),
        func.jsonb_object_agg(_minopt_cte.c.key, _minopt_cte.c.optimization_id).label("minimum_optimizations"),
    )
    .group_by(_minopt_cte.c.torsiondrive_id)
    .cte()
)


class TorsiondriveRecordORM(BaseRecordORM):
    """
    Table for storing torsiondrive calculations
    """

    __tablename__ = "torsiondrive_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(TorsiondriveSpecificationORM.id), nullable=False)
    specification = relationship(TorsiondriveSpecificationORM, lazy="selectin")

    initial_molecules = relationship(TorsiondriveInitialMoleculeORM, cascade="all, delete-orphan", passive_deletes=True)

    optimizations = relationship(
        TorsiondriveOptimizationORM,
        order_by=TorsiondriveOptimizationORM.position,
        collection_class=ordering_list("position"),
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    minimum_optimizations = column_property(
        select(_minopt_cte_agg.c.minimum_optimizations)
        .where(id == _minopt_cte_agg.c.torsiondrive_id)
        .scalar_subquery(),
        deferred=True,
    )

    __mapper_args__ = {
        "polymorphic_identity": "torsiondrive",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseRecordORM.model_dict(self, exclude)

        # Return initial molecule or just the ids, depending on what we have
        if "initial_molecules" in d:
            init_mol = d.pop("initial_molecules")
            d["initial_molecules_ids"] = [x["molecule_id"] for x in init_mol]
            if "molecule" in init_mol[0]:
                d["initial_molecules"] = [x["molecule"] for x in init_mol]

        return d

    @property
    def short_description(self) -> str:
        n_mol = len(self.initial_molecules)
        return f'{n_mol}*{self.initial_molecules[0].molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_torsiondrive_record_delete_base_tr
    AFTER DELETE ON torsiondrive_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(
    TorsiondriveRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql"))
)

record_direct_children_select = [
    select(BaseRecordORM.id.label("parent_id"), TorsiondriveOptimizationORM.optimization_id.label("child_id"))
    .join(TorsiondriveOptimizationORM, BaseRecordORM.id == TorsiondriveOptimizationORM.torsiondrive_id)
    .where(BaseRecordORM.record_type == "torsiondrive")
]

record_children_select = [
    *record_direct_children_select,
    select(BaseRecordORM.id.label("parent_id"), OptimizationTrajectoryORM.singlepoint_id.label("child_id"))
    .join(TorsiondriveOptimizationORM, BaseRecordORM.id == TorsiondriveOptimizationORM.torsiondrive_id)
    .join(
        OptimizationTrajectoryORM,
        OptimizationTrajectoryORM.optimization_id == TorsiondriveOptimizationORM.optimization_id,
    )
    .where(BaseRecordORM.record_type == "torsiondrive"),
]



"""Add external files table

Revision ID: 02afa97249c7
Revises: c13116948b54
Create Date: 2025-01-03 10:01:55.717905

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "02afa97249c7"
down_revision = "c13116948b54"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "external_file",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("created_on", sa.TIMESTAMP(), nullable=False),
        sa.Column("status", sa.Enum("available", "processing", name="externalfilestatusenum"), nullable=False),
        sa.Column("file_type", sa.Enum("dataset_attachment", name="externalfiletypeenum"), nullable=False),
        sa.Column("bucket", sa.String(), nullable=False),
        sa.Column("file_name", sa.String(), nullable=False),
        sa.Column("object_key", sa.String(), nullable=False),
        sa.Column("sha256sum", sa.String(), nullable=False),
        sa.Column("file_size", sa.BigInteger(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("provenance", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("external_file")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

from qcelemental.models import FailedOperation

from qcfractal.components.gridoptimization.testing_helpers import (
    submit_procedure_data as submit_go_procedure_data,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.torsiondrive.record_db_models import TorsiondriveRecordORM
from qcfractal.components.torsiondrive.testing_helpers import (
    submit_procedure_data as submit_td_procedure_data,
    generate_task_key as generate_td_task_key,
)
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.testing_helpers import run_service
from qcportal.managers import ManagerName
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


def test_service_socket_error(storage_socket: SQLAlchemySocket, session: Session, activated_manager_name: ManagerName):
    id_1, result_data_1 = submit_td_procedure_data(storage_socket, "td_H2O2_mopac_pm6", "test_tag", PriorityEnum.low)

    # Inject a failed computation
    failed_key = list(result_data_1.keys())[1]
    result_data_1[failed_key] = FailedOperation(
        error={"error_type": "test_error", "error_message": "this is just a test error"},
    )

    time_0 = now_at_utc()
    finished, n_optimizations = run_service(
        storage_socket, activated_manager_name, id_1, generate_td_task_key, result_data_1, 20
    )
    time_1 = now_at_utc()

    assert finished is True

    rec = session.get(TorsiondriveRecordORM, id_1)

    assert rec.status == RecordStatusEnum.error

    child_stat = storage_socket.records.torsiondrive.get_children_status(id_1, session=session)
    assert child_stat[RecordStatusEnum.error] == 1
    assert len(rec.compute_history) == 1
    assert len(rec.compute_history[-1].outputs) == 2  # stdout and error
    assert rec.compute_history[-1].status == RecordStatusEnum.error
    assert time_0 < rec.compute_history[-1].modified_on < time_1
    assert rec.service is not None

    err = rec.compute_history[-1].outputs["error"].get_output()
    assert "did not complete successfully" in err["error_message"]

    err_ids = [x.optimization_id for x in rec.optimizations if x.optimization_record.status == RecordStatusEnum.error]
    assert len(err_ids) == 1
    child_error_ids = storage_socket.records.torsiondrive.get_children_errors(id_1, session=session)
    assert len(child_error_ids) == 1

    assert child_error_ids[0] == err_ids[0]
    child_rec = session.get(BaseRecordORM, child_error_ids[0])
    assert child_rec.status == RecordStatusEnum.error


def test_service_socket_iterate_order(storage_socket: SQLAlchemySocket, session: Session):
    # TODO - Ugly
    max_active_services = storage_socket.services._max_active_services
    try:
        storage_socket.services._max_active_services = 1

        id_1, _ = submit_td_procedure_data(storage_socket, "td_H2O2_mopac_pm6", "*", PriorityEnum.normal)
        id_2, _ = submit_go_procedure_data(storage_socket, "go_H3NS_psi4_pbe", "*", PriorityEnum.high)

        with storage_socket.session_scope() as s:
            storage_socket.services.iterate_services(s)

        assert session.get(BaseRecordORM, id_1).status == RecordStatusEnum.waiting
        assert session.get(BaseRecordORM, id_2).status == RecordStatusEnum.running
    finally:
        storage_socket.services._max_active_services = max_active_services



"""Add progress description

Revision ID: c13116948b54
Revises: e798462e0c03
Create Date: 2025-01-07 10:35:23.654928

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c13116948b54"
down_revision = "e798462e0c03"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("internal_jobs", sa.Column("progress_description", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("internal_jobs", "progress_description")
    # ### end Alembic commands ###



"""delete old manybody

Revision ID: a5a701dc344d
Revises: 73b4838a6839
Create Date: 2024-06-11 15:51:11.380308

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a5a701dc344d"
down_revision = "73b4838a6839"
branch_labels = None
depends_on = None


def upgrade():

    conn = op.get_bind()

    res = conn.execute(sa.text("SELECT count(*) FROM base_record WHERE record_type = 'manybody'"))
    count = res.fetchone()[0]
    if count != 0:
        raise ValueError("Will not delete old manybody tables with existing data")

    res = conn.execute(sa.text("SELECT count(*) FROM manybody_record"))
    count = res.fetchone()[0]
    if count != 0:
        raise ValueError("Will not delete old manybody tables with existing data")

    res = conn.execute(sa.text("SELECT count(*) FROM manybody_dataset"))
    count = res.fetchone()[0]
    if count != 0:
        raise ValueError("Will not delete old manybody tables with existing data")

    op.execute(sa.text("DROP TABLE manybody_cluster CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_dataset CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_dataset_entry CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_dataset_record CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_dataset_specification CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_record CASCADE"))
    op.execute(sa.text("DROP TABLE manybody_specification CASCADE"))


def downgrade():
    raise NotImplementedError("Downgrade not supported")



from flask import jsonify, g
from flask_jwt_extended import (
    jwt_required,
    verify_jwt_in_request,
    get_jwt,
    get_jwt_identity,
)

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.auth_v1.blueprint import auth_v1
from qcfractal.flask_app.helpers import (
    get_all_endpoints,
    access_token_from_user,
    login_user_session,
    login_and_get_jwt,
    logout_user_session,
)


@auth_v1.route("/login", methods=["POST"])
def login():
    access_token, refresh_token = login_and_get_jwt(get_refresh_token=True)
    return jsonify(msg="Login succeeded!", access_token=access_token, refresh_token=refresh_token), 200


@auth_v1.route("/session_login", methods=["POST"])
def user_session_login():
    user_info = login_user_session()
    response = jsonify(
        msg="Login succeeded!",
        user_id=user_info.id,
        username=user_info.username,
        groups=user_info.groups,
        role=user_info.role,
    )
    return response, 200


@auth_v1.route("/session_logout", methods=["POST"])
def user_session_logout():
    logout_user_session()
    response = jsonify(msg="Logout successful!")
    return response, 200


@auth_v1.route("/refresh", methods=["POST"])
@jwt_required(refresh=True)
def refresh():
    user_id = get_jwt_identity()

    user_info = storage_socket.auth.verify(user_id)
    access_token = access_token_from_user(user_info)

    # For logging purposes (in the after_request_func)
    g.user_id = user_id

    ret = jsonify(access_token=access_token)
    return ret, 200


@auth_v1.route("/allowed", methods=["GET"])
def get_allowed_actions():
    all_endpoints = get_all_endpoints()
    all_actions = {"READ", "WRITE", "DELETE"}

    # JWT is optional
    verify_jwt_in_request(optional=True)

    username = get_jwt_identity()
    claims = get_jwt()
    policies = claims.get("permissions", ())

    subject = {"username": username}

    allowed = storage_socket.auth.allowed_actions(
        subject=subject, resources=all_endpoints, actions=all_actions, policies=policies
    )

    return jsonify(allowed), 200



"""rename tags on managers

Revision ID: 2c3c406a2750
Revises: 2abde51b4bc7
Create Date: 2025-03-21 10:36:19.961734

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "2c3c406a2750"
down_revision = "2abde51b4bc7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("compute_manager", "tags", new_column_name="compute_tags")

    op.drop_constraint("ck_compute_manager_tags_lower", "compute_manager")
    op.create_check_constraint(
        "ck_compute_manager_compute_tags_lower",
        "compute_manager",
        sa.column("compute_tags").cast(sa.TEXT) == sa.func.lower(sa.column("compute_tags").cast(sa.TEXT)),
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("compute_manager", "compute_tags", new_column_name="tags")

    op.drop_constraint("ck_compute_manager_tags_lower", "compute_manager")
    op.create_check_constraint(
        "ck_compute_manager_tags_lower",
        "compute_manager",
        sa.column("tags").cast(sa.TEXT) == sa.func.lower(sa.column("tags").cast(sa.TEXT)),
    )
    # ### end Alembic commands ###



from __future__ import annotations

import threading
import time
import uuid
from datetime import timedelta
from typing import TYPE_CHECKING

import pytest

from qcfractal.components.internal_jobs.db_models import InternalJobORM
from qcfractal.components.internal_jobs.socket import InternalJobSocket
from qcportal.internal_jobs import InternalJobStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


# Add in another function to the internal_jobs socket for testing
def dummy_internal_job(self, iterations: int, session, job_progress):
    assert session is not None
    assert job_progress is not None
    for i in range(iterations):
        time.sleep(1.0)
        job_progress.update_progress(100 * ((i + 1) / iterations))
        # print("Dummy internal job counter ", i)

        job_progress.raise_if_cancelled()

    return "Internal job finished"


# Add in another function to the internal_jobs socket for testing
# This one doesn't have session or job_progress
def dummy_internal_job_2(self, iterations: int):
    for i in range(iterations):
        time.sleep(1.0)
        # print("Dummy internal job counter ", i)

    return "Internal job finished"


setattr(InternalJobSocket, "dummy_job", dummy_internal_job)
setattr(InternalJobSocket, "dummy_job_2", dummy_internal_job_2)


def test_internal_jobs_socket_add_unique(storage_socket: SQLAlchemySocket):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=True
    )

    id_2 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=True
    )

    assert id_1 == id_2


def test_internal_jobs_socket_add_non_unique(storage_socket: SQLAlchemySocket):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    id_2 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    id_3 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    assert len({id_1, id_2, id_3}) == 3


@pytest.mark.parametrize("job_func", ("internal_jobs.dummy_job", "internal_jobs.dummy_job_2"))
def test_internal_jobs_socket_run(storage_socket: SQLAlchemySocket, session: Session, job_func: str):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), job_func, {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    time_0 = now_at_utc()
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(6)
    time_1 = now_at_utc()

    try:
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.running
        assert time_0 < job_1.last_updated < time_1

        if job_func == "internal_jobs.dummy_job":
            assert job_1.progress > 10
        else:
            assert job_1.progress == 0

        time.sleep(8)
        time_2 = now_at_utc()

        session.expire(job_1)
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.complete
        assert job_1.progress == 100
        assert job_1.result == "Internal job finished"
        assert time_1 < job_1.ended_date < time_2
        assert time_1 < job_1.last_updated < time_2

    finally:
        end_event.set()
        th.join()


@pytest.mark.parametrize("job_func", ("internal_jobs.dummy_job", "internal_jobs.dummy_job_2"))
def test_internal_jobs_socket_run_serial(storage_socket: SQLAlchemySocket, session: Session, job_func: str):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), job_func, {"iterations": 10}, None, unique_name=False, serial_group="test"
    )
    id_2 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), job_func, {"iterations": 10}, None, unique_name=False, serial_group="test"
    )
    id_3 = storage_socket.internal_jobs.add(
        "dummy_job",
        now_at_utc(),
        job_func,
        {"iterations": 10},
        None,
        unique_name=False,
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th1 = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th2 = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th3 = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th1.start()
    th2.start()
    th3.start()
    time.sleep(8)

    try:
        job_1 = session.get(InternalJobORM, id_1)
        job_2 = session.get(InternalJobORM, id_2)
        job_3 = session.get(InternalJobORM, id_3)
        assert job_1.status == InternalJobStatusEnum.running
        assert job_2.status == InternalJobStatusEnum.waiting
        assert job_3.status == InternalJobStatusEnum.running

    finally:
        end_event.set()
        th1.join()
        th2.join()
        th3.join()


def test_internal_jobs_socket_runnerstop(storage_socket: SQLAlchemySocket, session: Session):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 10}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(3)

    # Cancel/close the job runner
    end_event.set()
    th.join(20)
    assert not th.is_alive()

    job_1 = session.get(InternalJobORM, id_1)
    assert job_1.status == InternalJobStatusEnum.waiting
    assert job_1.progress == 0
    assert job_1.started_date is None
    assert job_1.last_updated is None
    assert job_1.runner_uuid is None

    return
    old_uuid = job_1.runner_uuid

    # Change uuid
    storage_socket.internal_jobs._uuid = str(uuid.uuid4())

    # Job is now running but orphaned. Should be picked up next time
    time.sleep(15)
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(30)

    try:
        session.expire(job_1)
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.complete
        assert job_1.runner_uuid != old_uuid
        assert job_1.progress == 100
        assert job_1.result == "Internal job finished"
    finally:
        end_event.set()
        th.join()


def test_internal_jobs_socket_recover(storage_socket: SQLAlchemySocket, session: Session):
    id_1 = storage_socket.internal_jobs.add(
        "dummy_job", now_at_utc(), "internal_jobs.dummy_job", {"iterations": 5}, None, unique_name=False
    )

    # Faster updates for testing
    storage_socket.internal_jobs._update_frequency = 1

    # Manually make it seem like it's running
    old_uuid = str(uuid.uuid4())
    job_1 = session.get(InternalJobORM, id_1)
    job_1.status = InternalJobStatusEnum.running
    job_1.progress = 10
    job_1.last_updated = now_at_utc() - timedelta(seconds=60)
    job_1.runner_uuid = old_uuid
    session.commit()

    session.expire(job_1)
    job_1 = session.get(InternalJobORM, id_1)
    assert job_1.status == InternalJobStatusEnum.running
    assert job_1.runner_uuid == old_uuid

    # Job is now running but orphaned. Should be picked up next time
    end_event = threading.Event()
    th = threading.Thread(target=storage_socket.internal_jobs.run_loop, args=(end_event,))
    th.start()
    time.sleep(10)

    try:
        session.expire(job_1)
        job_1 = session.get(InternalJobORM, id_1)
        assert job_1.status == InternalJobStatusEnum.complete
        assert job_1.runner_uuid != old_uuid
        assert job_1.progress == 100
        assert job_1.result == "Internal job finished"
    finally:
        end_event.set()
        th.join()



from __future__ import annotations

from sqlalchemy import (
    select,
    UniqueConstraint,
    Index,
    CheckConstraint,
    Column,
    Integer,
    ForeignKey,
    String,
    JSON,
    event,
    DDL,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, column_property

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.optimization.record_db_models import (
    OptimizationSpecificationORM,
    OptimizationRecordORM,
    OptimizationTrajectoryORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.db_socket import BaseORM


class GridoptimizationOptimizationORM(BaseORM):
    """
    Table for storing gridoptimization to optimization relationships
    """

    __tablename__ = "gridoptimization_optimization"

    gridoptimization_id = Column(
        Integer, ForeignKey("gridoptimization_record.id", ondelete="cascade"), primary_key=True
    )
    optimization_id = Column(Integer, ForeignKey("optimization_record.id"), nullable=False)
    key = Column(String, nullable=False, primary_key=True)

    energy = column_property(
        select(OptimizationRecordORM.energies[-1]).where(OptimizationRecordORM.id == optimization_id).scalar_subquery()
    )

    optimization_record = relationship(OptimizationRecordORM)

    __table_args__ = (Index("ix_gridoptimization_optimization_id", "optimization_id"),)

    _qcportal_model_excludes = ["gridoptimization_id"]


class GridoptimizationSpecificationORM(BaseORM):
    """
    Table for storing gridoptimization specifications
    """

    __tablename__ = "gridoptimization_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String(100), nullable=False)

    optimization_specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=False)
    optimization_specification = relationship(OptimizationSpecificationORM, lazy="joined")

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    __table_args__ = (
        UniqueConstraint(
            "specification_hash", "optimization_specification_id", name="ux_gridoptimization_specification_keys"
        ),
        Index("ix_gridoptimization_specification_program", "program"),
        Index("ix_gridoptimization_specification_optimization_specification_id", "optimization_specification_id"),
        # Enforce lowercase on some fields
        # This does not actually change the text to lowercase, but will fail to insert anything not lowercase
        # WARNING - these are not autodetected by alembic
        CheckConstraint("program = LOWER(program)", name="ck_gridoptimization_specification_program_lower"),
    )

    # TODO - protocols will be in model eventually
    _qcportal_model_excludes = [
        "id",
        "specification_hash",
        "optimization_specification_id",
        "protocols",
    ]

    @property
    def short_description(self) -> str:
        return f"{self.program}~{self.optimization_specification.short_description}"


class GridoptimizationRecordORM(BaseRecordORM):
    """
    Table for storing gridoptimization calculations
    """

    __tablename__ = "gridoptimization_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(GridoptimizationSpecificationORM.id), nullable=False)
    specification = relationship(GridoptimizationSpecificationORM, lazy="selectin")

    initial_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=False)
    initial_molecule = relationship(MoleculeORM, foreign_keys=initial_molecule_id)

    starting_molecule_id = Column(Integer, ForeignKey(MoleculeORM.id), nullable=True)
    starting_molecule = relationship(MoleculeORM, foreign_keys=starting_molecule_id)

    starting_grid = Column(JSON)  # tuple

    optimizations = relationship(GridoptimizationOptimizationORM, cascade="all, delete-orphan", passive_deletes=True)

    __mapper_args__ = {
        "polymorphic_identity": "gridoptimization",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    @property
    def short_description(self) -> str:
        return f'{self.initial_molecule.identifiers["molecular_formula"]} {self.specification.short_description}'


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL(
    """
    CREATE TRIGGER qca_gridoptimization_record_delete_base_tr
    AFTER DELETE ON gridoptimization_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """
)

event.listen(
    GridoptimizationRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql"))
)

record_direct_children_select = [
    select(BaseRecordORM.id.label("parent_id"), GridoptimizationOptimizationORM.optimization_id.label("child_id"))
    .join(GridoptimizationOptimizationORM, BaseRecordORM.id == GridoptimizationOptimizationORM.gridoptimization_id)
    .where(BaseRecordORM.record_type == "gridoptimization"),
]

record_children_select = [
    *record_direct_children_select,
    select(BaseRecordORM.id.label("parent_id"), OptimizationTrajectoryORM.singlepoint_id.label("child_id"))
    .join(GridoptimizationOptimizationORM, BaseRecordORM.id == GridoptimizationOptimizationORM.gridoptimization_id)
    .join(
        OptimizationTrajectoryORM,
        OptimizationTrajectoryORM.optimization_id == GridoptimizationOptimizationORM.optimization_id,
    )
    .where(BaseRecordORM.record_type == "gridoptimization"),
]



from __future__ import annotations

import threading
import time

from sqlalchemy import select

from qcfractal.components.optimization.record_db_models import OptimizationRecordORM
from qcfractal.components.optimization.testing_helpers import submit_procedure_data
from qcfractal.db_socket import SQLAlchemySocket
from qcfractal.db_socket.helpers import get_query_proj_options
from qcportal.molecules import Molecule


def test_dbsocket_helper_duplicate_insert(storage_socket: SQLAlchemySocket):
    # Tests that duplicate inserts are handled correctly if happening at the same time

    # Create another socket, so we can run the inserts in separate threads
    storage_socket_2 = SQLAlchemySocket(storage_socket.qcf_config)

    m1 = Molecule(symbols=["h", "h"], geometry=[0, 0, 0, 0, 0, 2])
    m2 = Molecule(symbols=["h", "h"], geometry=[0, 0, 0, 0, 0, 3])
    m3 = Molecule(symbols=["h", "h"], geometry=[0, 0, 0, 0, 0, 4])

    def _insert_molecules(s: SQLAlchemySocket):
        with s.session_scope() as session:
            # Add & flush, but don't commit
            r = s.molecules.add([m1, m2, m3], session=session)
            time.sleep(1)
            session.commit()
            return r

    # other thread inserts first
    t2 = threading.Thread(target=_insert_molecules, args=(storage_socket_2,))
    t2.start()
    time.sleep(0.25)

    # Now do ours
    _insert_molecules(storage_socket)
    t2.join()


def test_dbsocket_helper_proj(storage_socket: SQLAlchemySocket):
    empty_record_keys = {"id", "record_type"}

    record_id, _ = submit_procedure_data(storage_socket, "opt_psi4_methane_sometraj")

    with storage_socket.session_scope() as session:

        def run_proj(includes, excludes):
            stmt = select(OptimizationRecordORM).where(OptimizationRecordORM.id == record_id)
            query_opts = get_query_proj_options(OptimizationRecordORM, includes, excludes)
            stmt = stmt.options(*query_opts)
            d = session.execute(stmt).scalar_one().__dict__

            # remove sqlalchemy stuff
            return {k: v for k, v in d.items() if not k.startswith("_sa_")}

        # First, double check that some test columns are/aren't loaded by default)
        d = run_proj(None, None)
        assert "specification" in d
        assert "initial_molecule" not in d
        assert "trajectory" not in d
        assert "compute_history" not in d

        # default, no includes/excludes
        assert get_query_proj_options(OptimizationRecordORM) == []

        # explicit no includes/excludes
        assert get_query_proj_options(OptimizationRecordORM, None, None) == []

        # excludes is empty list
        assert get_query_proj_options(OptimizationRecordORM, None, []) == []

        # include only some columns
        d = run_proj(["initial_molecule_id"], None)
        assert "initial_molecule_id" in d
        assert "final_molecule_id" not in d
        assert "specification" not in d
        assert "specification_id" not in d

        # include only some relationships
        d = run_proj(["specification"], None)
        assert "initial_molecule_id" not in d
        assert "final_molecule_id" not in d
        assert "specification" in d
        assert "specification_id" not in d

        # include only some relationships not included by default (not allowed)
        d = run_proj(["initial_molecule", "compute_history"], None)
        assert "initial_molecule" not in d
        assert "compute_history" not in d
        assert "initial_molecule_id" not in d
        assert "final_molecule_id" not in d
        assert "specification" not in d
        assert "specification_id" not in d

        # include column + new rel + existing rel
        d = run_proj(["initial_molecule_id", "initial_molecule", "specification"], None)
        assert "initial_molecule" not in d  # new rel not allowed
        assert "initial_molecule_id" in d
        assert "final_molecule_id" not in d
        assert "specification" in d
        assert "specification_id" not in d

        #####################
        # Now some excludes
        #####################

        # Exclude columns and relationships that are loaded by default
        d = run_proj(None, ["initial_molecule_id", "specification"])
        assert "initial_molecule_id" not in d
        assert "specification" not in d
        assert "specification_id" in d

        d = run_proj(["*"], ["initial_molecule_id", "specification"])
        assert "initial_molecule_id" not in d
        assert "specification" not in d
        assert "specification_id" in d

        # Exclude columns and relationships that are not loaded by default
        assert get_query_proj_options(OptimizationRecordORM, None, ["initial_molecule"]) == []
        assert get_query_proj_options(OptimizationRecordORM, ["*"], ["initial_molecule"]) == []

        # Mixing includes & excludes (exclude overrides include)
        # Include/exclude column/relationships loaded by default
        d = run_proj(["specification", "initial_molecule_id"], ["specification", "initial_molecule_id"])
        assert d.keys() == empty_record_keys

        # Include/exclude column/relationships not loaded by default
        d = run_proj(["initial_molecule"], ["initial_molecule"])
        assert d.keys() == empty_record_keys

        # Handling non-existent columns/relationships
        d = run_proj(["not_a_column"], None)
        assert d.keys() == empty_record_keys

        d = run_proj(["not_a_column"], ["not_a_column"])
        assert d.keys() == empty_record_keys

        d = run_proj(["not_a_column", "initial_molecule_id"], ["not_a_column"])
        assert d.keys() == set(list(empty_record_keys) + ["initial_molecule_id"])

        assert get_query_proj_options(OptimizationRecordORM, None, ["not_a_column"]) == []

        # Empty queries
        # Note that id and record_type are always returned
        assert run_proj([], []).keys() == empty_record_keys

        assert run_proj(["initial_molecule_id"], ["initial_molecule_id"]).keys() == empty_record_keys

        # The only include is a relationship that can't be loaded
        assert run_proj(["initial_molecule"], None).keys() == empty_record_keys

        # Lastly, some complex examples
        d = run_proj(["specification", "initial_molecule_id", "compute_history"], ["specification_id"])
        assert "specification" in d
        assert "specification_id" not in d
        assert "initial_molecule_id" in d
        assert "status" not in d
        assert "compute_history" not in d



from __future__ import annotations

import pytest

from qcarchivetesting import test_users
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal import PortalClient


@pytest.fixture(scope="module")
def queryable_error_client(postgres_server, pytestconfig):
    pg_harness = postgres_server.get_new_harness("serverinfo_test_errors")
    encoding = pytestconfig.getoption("--client-encoding")

    with QCATestingSnowflake(pg_harness, encoding, create_users=True, enable_security=True, log_access=False) as server:
        # generate a bunch of test data
        storage_socket = server.get_storage_socket()

        admin_id = storage_socket.users.get("admin_user")["id"]
        read_id = storage_socket.users.get("read_user")["id"]

        with storage_socket.session_scope() as session:
            for i in range(20):
                for endpoint in ["molecules", "records", "wavefunctions", "managers"]:
                    for user_id in [read_id, admin_id]:
                        error = {
                            "error_text": f"ERROR_{i}_{endpoint}_{user_id}",
                            "user_id": user_id,
                            "request_path": "/api/v1/endpoint",
                            "request_headers": "fake_headers",
                            "request_body": "fake body",
                        }
                        storage_socket.serverinfo.save_error(error, session=session)

        yield server.client("admin_user", test_users["admin_user"]["pw"])


def test_serverinfo_client_query_error(queryable_error_client: PortalClient):
    # Query by user
    query_res = queryable_error_client.query_error_log(user="read_user")
    errors = list(query_res)
    assert len(errors) == 80
    assert all(x.user == "read_user" for x in errors)

    # Get a time
    test_time = errors[20].error_date
    query_res = queryable_error_client.query_error_log(user="read_user", before=test_time)
    errors = list(query_res)
    assert len(errors) == 60
    assert all(x.error_date <= test_time for x in errors)

    query_res = queryable_error_client.query_error_log(user="read_user", after=test_time)
    errors = list(query_res)
    assert len(errors) == 21
    assert all(x.error_date >= test_time for x in errors)

    # query by id
    ids = [errors[12].id, errors[15].id]
    query_res = queryable_error_client.query_error_log(error_id=ids)
    errors = list(query_res)
    assert len(errors) == 2
    assert all(x.id in ids for x in errors)


def test_serverinfo_client_query_error_empty_iter(queryable_error_client: PortalClient):
    query_res = queryable_error_client.query_error_log()
    assert len(query_res._current_batch) < queryable_error_client.api_limits["get_error_logs"]

    all_entries = list(query_res)
    assert len(all_entries) == 160


def test_serverinfo_client_query_error_limit(queryable_error_client: PortalClient):
    query_res = queryable_error_client.query_error_log(limit=38)
    all_entries = list(query_res)
    assert len(all_entries) == 38



"""Add repeat_delay to internal_jobs

Revision ID: e798462e0c03
Revises: 0587bb0220aa
Create Date: 2025-01-02 08:48:50.123907

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e798462e0c03"
down_revision = "0587bb0220aa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("internal_jobs", sa.Column("repeat_delay", sa.Integer(), nullable=True))

    # Remove old periodic tasks
    op.execute(
        """DELETE FROM internal_jobs WHERE status IN ('waiting', 'running') AND name IN (
                       'delete_old_internal_jobs',
                       'delete_old_access_log',
                       'iterate_services',
                       'geolocate_accesses',
                       'check_manager_heartbeats',
                       'update_geoip2_file'
                  )
               """
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("internal_jobs", "repeat_delay")
    # ### end Alembic commands ###



"""Update compute role for new endpoint

Revision ID: f8a7c273f18a
Revises: d03466436fec
Create Date: 2024-12-13 15:39:28.701912

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f8a7c273f18a"
down_revision = "d03466436fec"
branch_labels = None
depends_on = None


def upgrade():
    stmt = """
        UPDATE "role" SET permissions = jsonb_build_object(
                    'Statement', (permissions->'Statement')::jsonb || jsonb_build_object(
                               'Effect', 'Allow',
                               'Action', jsonb_build_array('READ'),
                               'Resource', '/compute/v1/information'
                )
            )::json
        WHERE rolename = 'compute'
        """

    op.execute(sa.text(stmt))


def downgrade():
    pass



from __future__ import annotations

from sqlalchemy import Column, String, Integer, ForeignKey, CheckConstraint, Index, UniqueConstraint, event, DDL, select
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION, JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.optimization.record_db_models import (
    OptimizationRecordORM,
    OptimizationSpecificationORM,
    OptimizationTrajectoryORM,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.record_db_models import SinglepointRecordORM, QCSpecificationORM
from qcfractal.db_socket import BaseORM


class ReactionComponentORM(BaseORM):
    """
    Table for storing reaction specifications
    """

    __tablename__ = "reaction_component"

    reaction_id = Column(Integer, ForeignKey("reaction_record.id", ondelete="cascade"), primary_key=True)
    molecule_id = Column(Integer, ForeignKey("molecule.id"), primary_key=True)
    coefficient = Column(DOUBLE_PRECISION, nullable=False)

    singlepoint_id = Column(Integer, ForeignKey(SinglepointRecordORM.id), nullable=True)
    optimization_id = Column(Integer, ForeignKey(OptimizationRecordORM.id), nullable=True)

    molecule = relationship(MoleculeORM)
    singlepoint_record = relationship(SinglepointRecordORM)
    optimization_record = relationship(OptimizationRecordORM)

    __table_args__ = (
        Index("ix_reaction_component_singlepoint_id", "singlepoint_id"),
        Index("ix_reaction_component_optimization_id", "optimization_id"),
    )

    _qcportal_model_excludes = ["reaction_id"]


class ReactionSpecificationORM(BaseORM):
    """
    Table for storing reaction specifications
    """

    __tablename__ = "reaction_specification"

    id = Column(Integer, primary_key=True)
    specification_hash = Column(String, nullable=False)

    program = Column(String, nullable=False)
    singlepoint_specification_id = Column(Integer, ForeignKey(QCSpecificationORM.id), nullable=True)
    optimization_specification_id = Column(Integer, ForeignKey(OptimizationSpecificationORM.id), nullable=True)

    keywords = Column(JSONB, nullable=False)
    protocols = Column(JSONB, nullable=False)

    singlepoint_specification = relationship(QCSpecificationORM, lazy="joined")
    optimization_specification = relationship(OptimizationSpecificationORM, lazy="joined")

    __table_args__ = (
        UniqueConstraint(
            "specification_hash",
            "singlepoint_specification_id",
            "optimization_specification_id",
            name="ux_reaction_specification_keys",
            postgresql_nulls_not_distinct=True,
        ),
        Index("ix_reaction_specification_program", "program"),
        Index("ix_reaction_specification_singlepoint_specification_id", "singlepoint_specification_id"),
        Index("ix_reaction_specification_optimization_specification_id", "optimization_specification_id"),
        CheckConstraint("program = LOWER(program)", name="ck_reaction_specification_program_lower"),
        CheckConstraint(
            "singlepoint_specification_id IS NOT NULL OR optimization_specification_id IS NOT NULL",
            name="ck_reaction_specification_specs",
        ),
    )

    # TODO - protocols will eventually be in the model
    _qcportal_model_excludes = [
        "id",
        "specification_hash",
        "singlepoint_specification_id",
        "optimization_specification_id",
        "protocols",
    ]

    @property
    def short_description(self) -> str:
        sp_desc = (
            self.singlepoint_specification.short_description
            if self.singlepoint_specification_id is not None
            else "(none)"
        )
        opt_desc = (
            self.optimization_specification.short_description
            if self.optimization_specification_id is not None
            else "(none)"
        )
        return f"{self.program}~[{sp_desc} | {opt_desc}]"


class ReactionRecordORM(BaseRecordORM):
    """
    Table for storing reaction calculations
    """

    __tablename__ = "reaction_record"

    id = Column(Integer, ForeignKey(BaseRecordORM.id, ondelete="cascade"), primary_key=True)

    specification_id = Column(Integer, ForeignKey(ReactionSpecificationORM.id), nullable=False)
    specification = relationship(ReactionSpecificationORM, lazy="selectin")

    total_energy = Column(DOUBLE_PRECISION, nullable=True)

    components = relationship(ReactionComponentORM, cascade="all, delete-orphan", passive_deletes=True)

    __mapper_args__ = {
        "polymorphic_identity": "reaction",
    }

    _qcportal_model_excludes = [*BaseRecordORM._qcportal_model_excludes, "specification_id"]

    @property
    def short_description(self) -> str:
        rxn_mols = ",".join(x.molecule.identifiers["molecular_formula"] for x in self.components)
        return f"[{rxn_mols}] {self.specification.short_description}"


# Delete base record if this record is deleted
_del_baserecord_trigger = DDL("""
    CREATE TRIGGER qca_reaction_record_delete_base_tr
    AFTER DELETE ON reaction_record
    FOR EACH ROW EXECUTE PROCEDURE qca_base_record_delete();
    """)

event.listen(ReactionRecordORM.__table__, "after_create", _del_baserecord_trigger.execute_if(dialect=("postgresql")))


record_direct_children_select = [
    select(BaseRecordORM.id.label("parent_id"), ReactionComponentORM.singlepoint_id.label("child_id"))
    .join(ReactionComponentORM, BaseRecordORM.id == ReactionComponentORM.reaction_id)
    .where(BaseRecordORM.record_type == "reaction", ReactionComponentORM.singlepoint_id.isnot(None)),
    select(BaseRecordORM.id.label("parent_id"), ReactionComponentORM.optimization_id.label("child_id"))
    .join(ReactionComponentORM, BaseRecordORM.id == ReactionComponentORM.reaction_id)
    .where(BaseRecordORM.record_type == "reaction", ReactionComponentORM.optimization_id.isnot(None)),
]

record_children_select = [
    *record_direct_children_select,
    select(BaseRecordORM.id.label("parent_id"), OptimizationTrajectoryORM.singlepoint_id.label("child_id"))
    .join(ReactionComponentORM, BaseRecordORM.id == ReactionComponentORM.reaction_id)
    .join(OptimizationTrajectoryORM, OptimizationTrajectoryORM.optimization_id == ReactionComponentORM.optimization_id)
    .where(BaseRecordORM.record_type == "reaction", ReactionComponentORM.optimization_id.isnot(None)),
]



from qcfractal.flask_app.compute_v1.blueprint import compute_v1
from qcfractal.flask_app.helpers import get_public_server_information
from qcfractal.flask_app.wrap_route import wrap_global_route


@compute_v1.route("/information", methods=["GET"])
@wrap_global_route("information", "read")
def get_information():
    return get_public_server_information()



from __future__ import annotations

import pytest
from qcelemental.models import Molecule

from qcportal import PortalClient


@pytest.fixture(scope="module")
def queryable_molecules_client(session_snowflake):
    client = session_snowflake.client()
    elements1 = ["h", "he", "li", "be", "b", "c", "n", "o", "f", "ne"]
    elements2 = ["na", "mg", "al", "si", "p", "s", "cl", "ar", "k", "ca"]

    all_mols = []
    for el1 in elements1:
        mols = []
        for el2 in elements2:
            for dist in range(2, 4, 1):
                m = Molecule(
                    symbols=[el1, el2],
                    geometry=[0, 0, 0, 0, 0, dist],
                    identifiers={
                        "smiles": f"madeupsmiles_{el1}_{el2}_{dist}",
                        "inchikey": f"madeupinchi_{el1}_{el2}_{dist}",
                    },
                )
                mols.append(m)

        meta, _ = client.add_molecules(mols)
        all_mols.extend(mols)
        assert meta.n_inserted == 20

    assert len(all_mols) == 200

    yield client
    session_snowflake.reset()


def test_molecules_client_query(queryable_molecules_client: PortalClient):
    def sort_molecules(m):
        return sorted(m, key=lambda x: x.get_hash())

    # Query by formula
    query_res = queryable_molecules_client.query_molecules(molecular_formula=["HNa", "CCl", "MgB"])
    mols = list(query_res)
    assert len(mols) == 6

    # Query by identifiers
    query_res = queryable_molecules_client.query_molecules(
        identifiers={"smiles": ["madeupsmiles_h_na_3", "madeupsmiles_c_s_2"]}
    )
    mols = list(query_res)
    assert len(mols) == 2

    query_res = queryable_molecules_client.query_molecules(
        identifiers={"inchikey": ["madeupinchi_c_cl_3", "madeupinchi_ne_ar_2"]}
    )
    mols = list(query_res)
    assert len(mols) == 2

    # Query by hash
    test_mols = mols[:3]
    test_hashes = [x.get_hash() for x in test_mols]
    query_res = queryable_molecules_client.query_molecules(molecule_hash=test_hashes)

    test_mols = sort_molecules(test_mols)
    res_mols = sort_molecules(query_res)
    assert test_mols == res_mols

    # Queries should be intersections
    query_res = queryable_molecules_client.query_molecules(
        molecular_formula=["HCl", "CS"], identifiers={"smiles": ["madeupsmiles_c_s_2"]}
    )
    mols = list(query_res)
    assert len(mols) == 1


def test_molecules_client_query_empty_iter(queryable_molecules_client: PortalClient):
    query_res = queryable_molecules_client.query_molecules()
    assert len(query_res._current_batch) < queryable_molecules_client.api_limits["get_molecules"]

    all_mols = list(query_res)
    assert len(all_mols) == 200


def test_molecules_client_query_limit(queryable_molecules_client: PortalClient):
    query_res = queryable_molecules_client.query_molecules(molecular_formula=["HCl", "CS"], limit=2)
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Limit that still requires batching
    query_res = queryable_molecules_client.query_molecules(limit=198)
    query_res_l = list(query_res)
    assert len(query_res_l) == 198



"""Task queue improvements

Revision ID: 6b24c66979ab
Revises: 49cfda6eb87e
Create Date: 2024-04-04 12:07:28.055915

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "6b24c66979ab"
down_revision = "49cfda6eb87e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("sort_date", postgresql.TIMESTAMP(timezone=True), nullable=True))
    op.create_index(
        "ix_task_queue_sort",
        "task_queue",
        [sa.text("priority DESC"), sa.text("sort_date ASC"), sa.text("id ASC")],
        unique=False,
    )

    op.execute(
        sa.text(
            """
        UPDATE task_queue
        SET sort_date = br.created_on
        FROM base_record br
        WHERE task_queue.record_id = br.id
        """
        )
    )

    op.alter_column("task_queue", "sort_date", nullable=False)

    op.drop_index("ix_task_queue_required_programs", table_name="task_queue")
    op.create_index(
        "ix_task_queue_required_programs", "task_queue", [sa.text("required_programs")], postgresql_using="gin"
    )

    op.drop_index("ix_base_record_created_on", table_name="base_record")
    op.create_index("ix_base_record_created_on", "base_record", ["created_on"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_base_record_created_on", table_name="base_record")
    op.create_index("ix_base_record_created_on", "base_record", ["created_on"], postgresql_using="brin")

    op.drop_index("ix_task_queue_required_programs", table_name="task_queue")
    op.create_index("ix_task_queue_required_programs", "task_queue", [sa.text("required_programs")])

    op.drop_index("ix_task_queue_sort", table_name="task_queue")
    op.drop_column("task_queue", "sort_date")

    # ### end Alembic commands ###



from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Optional

import pytest

from qcarchivetesting import load_molecule_data
from qcfractal.components.manybody.record_db_models import ManybodyRecordORM
from qcportal.manybody import ManybodySpecification
from qcportal.record_models import RecordStatusEnum, PriorityEnum
from qcportal.singlepoint import QCSpecification
from qcportal.utils import now_at_utc
from .testing_helpers import compare_manybody_specs, test_specs, submit_procedure_data, run_procedure_data

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcportal import PortalClient


def test_manybody_client_tag_priority(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    for tag, priority in itertools.product(["*", "tag99"], list(PriorityEnum)):
        sp_spec = QCSpecification(
            program="prog",
            driver="energy",
            method="hf",
            basis="sto-3g",
            keywords={"tag_priority": [tag, priority]},
        )

        meta1, id1 = snowflake_client.add_manybodys(
            [water],
            "qcmanybody",
            bsse_correction=["nocp"],
            levels={1: sp_spec},
            keywords={"return_total_data": True},
            compute_tag=tag,
            compute_priority=priority,
        )

        assert meta1.n_inserted == 1
        rec = snowflake_client.get_records(id1, include=["service"])
        assert rec[0].service.compute_tag == tag
        assert rec[0].service.compute_priority == priority


@pytest.mark.parametrize("spec", test_specs)
def test_manybody_client_add_get(
    submitter_client: PortalClient,
    spec: ManybodySpecification,
):
    water2 = load_molecule_data("water_dimer_minima")
    water4 = load_molecule_data("water_stacked")

    time_0 = now_at_utc()
    meta1, id1 = submitter_client.add_manybodys(
        [water2, water4],
        spec.program,
        spec.levels,
        spec.bsse_correction,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )
    time_1 = now_at_utc()
    assert meta1.success

    recs = submitter_client.get_manybodys(id1, include=["service", "clusters", "initial_molecule"])
    assert len(recs) == 2

    for r in recs:
        assert r.record_type == "manybody"
        assert r.record_type == "manybody"
        assert compare_manybody_specs(spec, r.specification)

        assert r.status == RecordStatusEnum.waiting
        assert r.children_status == {}

        assert r.service.compute_tag == "tag1"
        assert r.service.compute_priority == PriorityEnum.low

        assert r.creator_user == submitter_client.username

        assert time_0 < r.created_on < time_1
        assert time_0 < r.modified_on < time_1

    assert recs[0].initial_molecule.get_hash() == water2.get_hash()
    assert recs[1].initial_molecule.get_hash() == water4.get_hash()


@pytest.mark.parametrize("spec", test_specs)
@pytest.mark.parametrize("find_existing", [True, False])
def test_manybody_client_add_duplicate(
    submitter_client: PortalClient, spec: ManybodySpecification, find_existing: bool
):
    water2 = load_molecule_data("water_dimer_minima")
    water4 = load_molecule_data("water_stacked")
    all_mols = [water2, water4]

    meta, id = submitter_client.add_manybodys(
        all_mols,
        spec.program,
        spec.levels,
        spec.bsse_correction,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=True,
    )
    assert meta.success
    assert meta.n_inserted == len(all_mols)

    meta, id2 = submitter_client.add_manybodys(
        all_mols,
        spec.program,
        spec.levels,
        spec.bsse_correction,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
        find_existing=find_existing,
    )

    if find_existing:
        assert meta.n_existing == len(all_mols)
        assert meta.n_inserted == 0
        assert id == id2
    else:
        assert meta.n_existing == 0
        assert meta.n_inserted == len(all_mols)
        assert set(id).isdisjoint(id2)


def test_manybody_client_add_existing_molecule(snowflake_client: PortalClient):
    spec = test_specs[0]

    mol1 = load_molecule_data("water_dimer_minima")
    mol2 = load_molecule_data("water_stacked")

    # Add a molecule separately
    _, mol_ids = snowflake_client.add_molecules([mol1])

    # Now add records
    meta1, id1 = snowflake_client.add_manybodys(
        [mol1, mol2, mol1],
        spec.program,
        spec.levels,
        spec.bsse_correction,
        keywords=spec.keywords,
        compute_tag="tag1",
        compute_priority=PriorityEnum.low,
    )

    assert meta1.success

    assert meta1.success
    assert meta1.n_inserted == 2
    assert meta1.n_existing == 1

    recs = snowflake_client.get_manybodys(id1, include=["initial_molecule"])
    assert len(recs) == 3
    assert recs[0].id == recs[2].id
    assert recs[0].id != recs[1].id

    assert recs[0].initial_molecule.get_hash() == mol1.get_hash()
    assert recs[1].initial_molecule.get_hash() == mol2.get_hash()


def test_manybody_client_delete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    mb_id = run_procedure_data(storage_socket, activated_manager_name, "mb_cp_he4_psi4_mp2")

    with storage_socket.session_scope() as session:
        rec = session.get(ManybodyRecordORM, mb_id)
        child_ids = [x.singlepoint_id for x in rec.clusters]

    meta = snowflake_client.delete_records(mb_id, soft_delete=True, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.complete for x in child_recs)
    mb_rec = snowflake_client.get_records(mb_id)
    assert mb_rec.children_status == {RecordStatusEnum.complete: len(child_ids)}

    snowflake_client.undelete_records(mb_id)

    meta = snowflake_client.delete_records(mb_id, soft_delete=True, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(set(child_ids))

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x.status == RecordStatusEnum.deleted for x in child_recs)
    mb_rec = snowflake_client.get_records(mb_id)
    assert mb_rec.children_status == {RecordStatusEnum.deleted: len(child_ids)}

    meta = snowflake_client.delete_records(mb_id, soft_delete=False, delete_children=True)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == len(set(child_ids))

    recs = snowflake_client.get_manybodys(mb_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is None for x in child_recs)

    # DB should be pretty empty now
    query_res = snowflake_client.query_records()
    query_res_l = list(query_res)
    assert len(query_res_l) == 0


def test_manybody_client_harddelete_nochildren(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    mb_id = run_procedure_data(storage_socket, activated_manager_name, "mb_cp_he4_psi4_mp2")

    with storage_socket.session_scope() as session:
        rec = session.get(ManybodyRecordORM, mb_id)
        child_ids = [x.singlepoint_id for x in rec.clusters]

    meta = snowflake_client.delete_records(mb_id, soft_delete=False, delete_children=False)
    assert meta.success
    assert meta.deleted_idx == [0]
    assert meta.n_children_deleted == 0

    recs = snowflake_client.get_manybodys(mb_id, missing_ok=True)
    assert recs is None

    child_recs = snowflake_client.get_records(child_ids, missing_ok=True)
    assert all(x is not None for x in child_recs)


def test_manybody_client_delete_opt_inuse(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    mb_id = run_procedure_data(storage_socket, activated_manager_name, "mb_cp_he4_psi4_mp2")

    with storage_socket.session_scope() as session:
        rec = session.get(ManybodyRecordORM, mb_id)
        child_ids = [x.singlepoint_id for x in rec.clusters]

    meta = snowflake_client.delete_records(child_ids[0], soft_delete=False)
    assert meta.success is False
    assert meta.error_idx == [0]

    ch_rec = snowflake_client.get_records(child_ids[0])
    assert ch_rec is not None


def test_manybody_client_query(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    id_1, _ = submit_procedure_data(storage_socket, "mb_cp_he4_psi4_mp2")
    id_2, _ = submit_procedure_data(storage_socket, "mb_all_he4_psi4_multiss")

    all_mbs = snowflake_client.get_manybodys([id_1, id_2])
    mol_ids = [x.initial_molecule_id for x in all_mbs]

    query_res = snowflake_client.query_manybodys(program=["qcmanybody"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_manybodys(program=["nothing"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_manybodys(initial_molecule_id=[9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_manybodys(initial_molecule_id=[mol_ids[0], 9999])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # query for qc program
    query_res = snowflake_client.query_manybodys(qc_program=["Psi4"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_manybodys(qc_program=["abc"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for basis
    query_res = snowflake_client.query_manybodys(qc_basis=["DEF2-tzvp"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_manybodys(qc_basis=["6-31g*"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_manybodys(qc_basis=["6-31g"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    query_res = snowflake_client.query_manybodys(qc_basis=[None])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    query_res = snowflake_client.query_manybodys(qc_basis=[""])
    query_res_l = list(query_res)
    assert len(query_res_l) == 0

    # query for method
    query_res = snowflake_client.query_manybodys(qc_method=["hf"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 1

    query_res = snowflake_client.query_manybodys(qc_method=["mp2"])
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Query by default returns everything
    query_res = snowflake_client.query_manybodys()
    query_res_l = list(query_res)
    assert len(query_res_l) == 2

    # Query by default (with a limit)
    query_res = snowflake_client.query_manybodys(limit=1)
    query_res_l = list(query_res)
    assert len(query_res_l) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcarchivetesting.helpers import test_users, read_record_data
from qcfractal.components.singlepoint.testing_helpers import load_procedure_data, run_procedure_data
from qcportal import PortalRequestError
from qcportal.compare_records import compare_records
from qcportal.record_models import PriorityEnum, RecordStatusEnum, record_from_dict
from qcportal.singlepoint import SinglepointInput

if TYPE_CHECKING:
    from qcportal import PortalClient
    from qcarchivetesting.testing_classes import QCATestingSnowflake


test_inp_1 = SinglepointInput(
    molecule={"symbols": ["h", "h"], "geometry": [0, 0, 0, 0, 0, 2]},
    specification={
        "program": "psi4",
        "driver": "energy",
        "method": "hf",
        "basis": "sto-3g",
    },
)


def test_project_client_add_get(submitter_client: PortalClient):
    proj = submitter_client.add_project(
        "test project",
        "Test Description",
        "a Tagline",
        ["tag1", "tag2"],
        "test_compute_tag",
        PriorityEnum.low,
        {"meta_key_1": "meta_value_1"},
    )

    assert proj.name == "test project"
    assert proj.description == "Test Description"
    assert proj.tagline == "a Tagline"
    assert proj.tags == ["tag1", "tag2"]
    assert proj.default_compute_tag == "test_compute_tag"
    assert proj.default_compute_priority == PriorityEnum.low
    assert proj.extras == {"meta_key_1": "meta_value_1"}

    assert proj.owner_user == submitter_client.username

    # case insensitive
    proj2 = submitter_client.get_project("TEST PrOJECT")
    assert proj2.id == proj.id

    plist = submitter_client.list_projects()
    assert plist[0]["id"] == proj.id
    assert plist[0]["project_name"] == proj.name
    assert plist[0]["tagline"] == proj.tagline
    assert plist[0]["tags"] == proj.tags
    assert plist[0]["owner_user"] == proj.owner_user


def test_project_client_add_get_records_datasets(snowflake_client: PortalClient):
    proj = snowflake_client.add_project(
        "test project",
        default_compute_tag="test_compute_tag",
        default_compute_priority=PriorityEnum.low,
    )

    r = proj.add_record("test_record", test_inp_1)
    ds1 = proj.add_dataset("singlepoint", "test singlepoint dataset")
    ds2 = proj.add_dataset("optimization", "test optimization dataset")

    r_test = proj.get_record("test_record")
    ds1_test = proj.get_dataset("test singlepoint dataset")
    ds2_test = proj.get_dataset("test optimization dataset")

    assert r_test.id == r.id
    assert r.task.compute_tag == "test_compute_tag"
    assert r.task.compute_priority == PriorityEnum.low

    assert ds1_test.id == ds1.id
    assert ds1.default_compute_tag == "test_compute_tag"
    assert ds1_test.default_compute_priority == PriorityEnum.low

    assert ds2_test.id == ds2.id
    assert ds2_test.default_compute_tag == "test_compute_tag"
    assert ds2_test.default_compute_priority == PriorityEnum.low

    plist = snowflake_client.list_projects()
    assert plist[0]["id"] == proj.id
    assert plist[0]["project_name"] == proj.name
    assert plist[0]["record_count"] == 1
    assert plist[0]["dataset_count"] == 2


def test_project_client_link_records_datasets(snowflake_client: PortalClient):

    # Add these directly to the server (not part of the project)
    meta, rids = snowflake_client.add_singlepoints(
        molecules=[test_inp_1.molecule],
        program=test_inp_1.specification.program,
        driver=test_inp_1.specification.driver,
        method=test_inp_1.specification.method,
        basis=test_inp_1.specification.basis,
        keywords=test_inp_1.specification.keywords,
        compute_tag="test_compute_tag",
        compute_priority=PriorityEnum.low,
    )

    assert meta.success

    ds1 = snowflake_client.add_dataset(
        "singlepoint", "test singlepoint dataset", description="description", tagline="tagline", tags=["tag2"]
    )
    ds2 = snowflake_client.add_dataset(
        "optimization", "test optimization dataset", description="description 2", tagline="tagline 2", tags=["tag4"]
    )

    proj = snowflake_client.add_project(
        "test project",
        default_compute_tag="test_compute_tag",
        default_compute_priority=PriorityEnum.low,
    )

    linked_ds1 = proj.link_dataset(ds1.id)

    # Changing the name
    linked_ds2 = proj.link_dataset(
        ds2.id, name="new name", description="new description", tagline="new tagline", tags=["tag3"]
    )

    linked_r = proj.link_record(rids[0], "record_name", "description", tags=["tag1"])

    # Refetch to make sure it's changed on the server
    linked_ds1 = proj.get_dataset("test singlepoint dataset")
    linked_ds2 = proj.get_dataset("new name")
    linked_r = proj.get_record("record_name")

    assert linked_r.id == rids[0]
    assert linked_r.name == "record_name"
    assert linked_r.description == "description"
    assert linked_r.tags == ["tag1"]

    assert linked_ds1.id == ds1.id
    assert linked_ds1.name == "test singlepoint dataset"
    assert linked_ds1.description == "description"
    assert linked_ds1.tagline == "tagline"
    assert linked_ds1.tags == ["tag2"]

    assert linked_ds2.id == ds2.id
    assert linked_ds2.name == "new name"
    assert linked_ds2.description == "new description"
    assert linked_ds2.tagline == "new tagline"
    assert linked_ds2.tags == ["tag3"]

    # Can't link again
    with pytest.raises(PortalRequestError, match="Dataset.*already linked"):
        proj.link_dataset(ds1.id)

    with pytest.raises(PortalRequestError, match="Record.*already linked"):
        proj.link_record(rids[0], "record_name", "description", tags=["tag1"])

    # Unlink stuff
    proj.unlink_datasets("test singlepoint dataset")
    proj.unlink_records("record_name")
    proj.unlink_datasets("new name")

    assert len(proj.dataset_metadata) == 0
    assert len(proj.record_metadata) == 0
    proj.fetch_dataset_metadata()
    proj.fetch_record_metadata()
    assert len(proj.dataset_metadata) == 0
    assert len(proj.record_metadata) == 0


def test_project_client_delete(snowflake_client: PortalClient):
    proj = snowflake_client.add_project("test project")

    ds1 = proj.add_dataset("singlepoint", "test singlepoint dataset")
    ds2 = proj.add_dataset("optimization", "test optimization dataset")

    r = proj.add_record("test_record", test_inp_1)

    snowflake_client.delete_project(proj.id)

    with pytest.raises(PortalRequestError, match="Could not find project"):
        snowflake_client.get_project("test project")

    with pytest.raises(PortalRequestError, match="Could not find project"):
        snowflake_client.get_project_by_id(proj.id)

    r_test = snowflake_client.get_records(r.id)
    ds1_test = snowflake_client.get_dataset_by_id(ds1.id)
    ds2_test = snowflake_client.get_dataset_by_id(ds2.id)

    assert r_test.id == r.id
    assert ds1_test.id == ds1.id and ds1_test.name == "test singlepoint dataset"
    assert ds2_test.id == ds2.id and ds2_test.name == "test optimization dataset"


def test_project_client_delete_with_records(snowflake_client: PortalClient):
    proj = snowflake_client.add_project("test project")

    ds1 = proj.add_dataset("singlepoint", "test singlepoint dataset")
    ds2 = proj.add_dataset("optimization", "test optimization dataset")
    r = proj.add_record("test_record", test_inp_1)

    snowflake_client.delete_project(proj.id, delete_records=True, delete_datasets=False)

    with pytest.raises(PortalRequestError, match="Could not find project"):
        snowflake_client.get_project("test project")

    r_test = snowflake_client.get_records(r.id, missing_ok=True)
    ds1_test = snowflake_client.get_dataset_by_id(ds1.id)
    ds2_test = snowflake_client.get_dataset_by_id(ds2.id)

    assert r_test is None
    assert ds1_test.id == ds1.id and ds1_test.name == "test singlepoint dataset"
    assert ds2_test.id == ds2.id and ds2_test.name == "test optimization dataset"


def test_project_client_delete_with_datasets(snowflake_client: PortalClient):
    proj = snowflake_client.add_project("test project")

    ds1 = proj.add_dataset("singlepoint", "test singlepoint dataset")
    ds2 = proj.add_dataset("optimization", "test optimization dataset")
    r = proj.add_record("test_record", test_inp_1)

    snowflake_client.delete_project(proj.id, delete_records=False, delete_datasets=True)

    with pytest.raises(PortalRequestError, match="Could not find project"):
        snowflake_client.get_project_by_id(proj.id)

    r_test = snowflake_client.get_records(r.id)
    assert r_test.id == r.id

    with pytest.raises(PortalRequestError, match="Could not find dataset"):
        snowflake_client.get_dataset_by_id(ds1.id)
    with pytest.raises(PortalRequestError, match="Could not find dataset"):
        snowflake_client.get_dataset_by_id(ds2.id)


def test_project_client_status(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()

    proj = snowflake_client.add_project(
        "test project",
        default_compute_tag="test_compute_tag",
        default_compute_priority=PriorityEnum.low,
    )

    ds = proj.add_dataset("singlepoint", "test singlepoint dataset")

    input_spec, molecule, _ = load_procedure_data("sp_psi4_peroxide_energy_wfn")
    run_procedure_data(storage_socket, activated_manager_name, "sp_psi4_peroxide_energy_wfn")

    # Add this as a part of the dataset
    ds.add_specification("spec_1", input_spec)
    ds.add_entry(name="test_molecule", molecule=molecule)
    ds.add_entry(name="test_molecule_2", molecule=molecule, additional_keywords={"maxiter": 999})
    ds.submit()

    # Also add a lone record
    r = proj.add_record("test_record", test_inp_1)

    status = proj.status()
    assert len(status["records"]) == 1
    assert status["records"][RecordStatusEnum.waiting] == 1
    assert status["datasets"] == {RecordStatusEnum.complete: 1, RecordStatusEnum.waiting: 1}


def test_project_client_add_duplicates(snowflake_client: PortalClient):
    proj = snowflake_client.add_project("test project")

    proj.add_record("test_record", test_inp_1)
    proj.add_dataset("singlepoint", "test singlepoint dataset")
    proj.add_dataset("optimization", "test optimization dataset")

    with pytest.raises(PortalRequestError, match="Record 'test_record' already exists in project"):
        proj.add_record("test_record", test_inp_1)

    with pytest.raises(PortalRequestError, match="Dataset 'test optimization dataset' already exists in project"):
        proj.add_dataset("optimization", "test optimization dataset")

    with pytest.raises(PortalRequestError, match="Dataset 'test optimization dataset' already exists in project"):
        proj.add_dataset("optimization", "test optimization dataset")


def test_project_client_import_records(secure_snowflake: QCATestingSnowflake):

    client = secure_snowflake.client("submit_user", test_users["submit_user"]["pw"])
    proj = client.add_project("test project")

    test_files = [
        "go_error_17761737",
        "go_H2O2_psi4_b3lyp",
        "mb_all_he4_psi4_multiss",
        "neb_error_68251163",
        "neb_HCN_psi4_b3lyp_opt3",
        "opt_error_118868739",
        "opt_psi4_benzene",
        "rxn_error_118326390",
        "rxn_H2O_psi4_mp2_optsp",
        "sp_error_119608646",
        "sp_psi4_benzene_energy_1",
        "sp_psi4_h2_b3lyp_nativefiles",
        "sp_psi4_peroxide_energy_wfn",
        "td_error_137174663",
        "td_H2O2_mopac_pm6",
    ]

    test_data = {}
    id_map = {}

    for t_name in test_files:
        d = read_record_data(t_name)
        r = record_from_dict(d)
        test_data[t_name] = r

        tag = t_name.split("_")[0]
        server_r = proj.import_record(t_name, r, description=f"Record {t_name}", tags=["imported", tag])
        id_map[t_name] = server_r.id

    # Check that they are there
    proj = client.get_project("test project")
    for rm in proj.record_metadata:
        assert rm.tags == ["imported", rm.name.split("_")[0]]
        server_r = proj.get_record(rm.record_id, include=["**"])
        server_r.fetch_children(include=["**"])

        test_r = test_data[rm.name]
        compare_records(server_r, test_r)
        assert server_r.creator_user == "submit_user"



from typing import List

from flask import current_app, g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import LimitExceededError
from qcportal.reaction import (
    ReactionDatasetSpecification,
    ReactionDatasetNewEntry,
    ReactionAddBody,
    ReactionQueryFilters,
)
from qcportal.utils import calculate_limit


#####################
# Record
#####################


@api_v1.route("/records/reaction/bulkCreate", methods=["POST"])
@wrap_global_route("records", "add")
def add_reaction_records_v1(body_data: ReactionAddBody):
    limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.add_records
    if len(body_data.stoichiometries) > limit:
        raise LimitExceededError(f"Cannot add {len(body_data.stoichiometries)} reaction records - limit is {limit}")

    return storage_socket.records.reaction.add(
        stoichiometries=body_data.stoichiometries,
        rxn_spec=body_data.specification,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/records/reaction/<int:record_id>/components", methods=["GET"])
@wrap_global_route("records", "read")
def get_reaction_components_v1(record_id: int):
    return storage_socket.records.reaction.get_components(record_id)


@api_v1.route("/records/reaction/query", methods=["POST"])
@wrap_global_route("records", "read")
def query_reaction_v1(body_data: ReactionQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_records
    body_data.limit = calculate_limit(max_limit, body_data.limit)

    return storage_socket.records.reaction.query(body_data)


#####################
# Dataset
#####################


@api_v1.route("/datasets/reaction/<int:dataset_id>/specifications", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_reaction_dataset_specifications_v1(dataset_id: int, body_data: List[ReactionDatasetSpecification]):
    return storage_socket.datasets.reaction.add_specifications(dataset_id, body_data)


@api_v1.route("/datasets/reaction/<int:dataset_id>/entries/bulkCreate", methods=["POST"])
@wrap_global_route("datasets", "modify")
def add_reaction_dataset_entries_v1(dataset_id: int, body_data: List[ReactionDatasetNewEntry]):
    return storage_socket.datasets.reaction.add_entries(dataset_id, new_entries=body_data)


@api_v1.route("/datasets/reaction/<int:dataset_id>/background_add_entries", methods=["POST"])
@wrap_global_route("datasets", "modify")
def background_add_reaction_dataset_entries_v1(dataset_id: int, body_data: List[ReactionDatasetNewEntry]):
    return storage_socket.datasets.reaction.background_add_entries(dataset_id, new_entries=body_data)



"""change session data column type

Revision ID: e15456ce68e2
Revises: 8738641e0b90
Create Date: 2025-04-06 09:17:25.122270

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "e15456ce68e2"
down_revision = "8738641e0b90"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("DELETE FROM user_session")
    op.add_column("user_session", sa.Column("session_key", sa.String(), nullable=False))

    op.drop_column("user_session", "session_data")
    op.add_column("user_session", sa.Column("session_data", postgresql.JSONB(astext_type=sa.Text()), nullable=False))

    op.drop_constraint("ux_user_session_session_id", "user_session", type_="unique")
    op.create_unique_constraint("ux_user_session_session_key", "user_session", ["session_key"])
    op.drop_column("user_session", "session_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("user_session", sa.Column("session_id", sa.VARCHAR(), autoincrement=False, nullable=False))
    op.drop_constraint("ux_user_session_session_key", "user_session", type_="unique")
    op.create_unique_constraint("ux_user_session_session_id", "user_session", ["session_id"])

    op.drop_column("user_session", "session_data")
    op.add_column("user_session", sa.Column("session_data", postgresql.BYTEA(), nullable=False))

    op.drop_column("user_session", "session_key")
    # ### end Alembic commands ###



from __future__ import annotations

from sqlalchemy import Column, Integer, ForeignKey, String, ForeignKeyConstraint, Index, select
from sqlalchemy.dialects.postgresql import DOUBLE_PRECISION, JSONB
from sqlalchemy.orm import relationship

from qcfractal.components.dataset_db_models import BaseDatasetORM
from qcfractal.components.molecules.db_models import MoleculeORM
from qcfractal.components.reaction.record_db_models import ReactionRecordORM, ReactionSpecificationORM
from qcfractal.db_socket import BaseORM


class ReactionDatasetStoichiometryORM(BaseORM):
    __tablename__ = "reaction_dataset_stoichiometry"

    dataset_id = Column(Integer, ForeignKey("reaction_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)

    molecule_id = Column(Integer, ForeignKey("molecule.id"), primary_key=True)
    coefficient = Column(DOUBLE_PRECISION, nullable=False)

    molecule = relationship(MoleculeORM, lazy="joined")

    __table_args__ = (
        Index("ix_reaction_dataset_stoichiometry_dataset_id", "dataset_id"),
        Index("ix_reaction_dataset_stoichiometry_entry_name", "entry_name"),
        Index("ix_reaction_dataset_stoichiometry_molecule_id", "molecule_id"),
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["reaction_dataset_entry.dataset_id", "reaction_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
    )

    _qcportal_model_excludes = ["dataset_id", "molecule_id"]


class ReactionDatasetEntryORM(BaseORM):
    __tablename__ = "reaction_dataset_entry"

    dataset_id = Column(Integer, ForeignKey("reaction_dataset.id", ondelete="cascade"), primary_key=True)

    name = Column(String, primary_key=True)
    comment = Column(String)

    additional_keywords = Column(JSONB, nullable=True)
    attributes = Column(JSONB, nullable=False)

    stoichiometries = relationship(
        ReactionDatasetStoichiometryORM, lazy="selectin", cascade="all, delete-orphan", passive_deletes=True
    )

    __table_args__ = (
        Index("ix_reaction_dataset_entry_dataset_id", "dataset_id"),
        Index("ix_reaction_dataset_entry_name", "name"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class ReactionDatasetSpecificationORM(BaseORM):
    __tablename__ = "reaction_dataset_specification"

    dataset_id = Column(Integer, ForeignKey("reaction_dataset.id", ondelete="cascade"), primary_key=True)
    name = Column(String, primary_key=True)
    description = Column(String, nullable=True)
    specification_id = Column(Integer, ForeignKey(ReactionSpecificationORM.id), nullable=False)

    specification = relationship(ReactionSpecificationORM, uselist=False)

    __table_args__ = (
        Index("ix_reaction_dataset_specification_dataset_id", "dataset_id"),
        Index("ix_reaction_dataset_specification_name", "name"),
        Index("ix_reaction_dataset_specification_specification_id", "specification_id"),
    )

    _qcportal_model_excludes = ["dataset_id", "specification_id"]


class ReactionDatasetRecordItemORM(BaseORM):
    __tablename__ = "reaction_dataset_record"

    dataset_id = Column(Integer, ForeignKey("reaction_dataset.id", ondelete="cascade"), primary_key=True)
    entry_name = Column(String, primary_key=True)
    specification_name = Column(String, primary_key=True)
    record_id = Column(Integer, ForeignKey(ReactionRecordORM.id), nullable=False)

    record = relationship(ReactionRecordORM)

    __table_args__ = (
        ForeignKeyConstraint(
            ["dataset_id", "entry_name"],
            ["reaction_dataset_entry.dataset_id", "reaction_dataset_entry.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        ForeignKeyConstraint(
            ["dataset_id", "specification_name"],
            ["reaction_dataset_specification.dataset_id", "reaction_dataset_specification.name"],
            ondelete="cascade",
            onupdate="cascade",
        ),
        Index("ix_reaction_dataset_record_record_id", "record_id"),
    )

    _qcportal_model_excludes = ["dataset_id"]


class ReactionDatasetORM(BaseDatasetORM):
    __tablename__ = "reaction_dataset"

    id = Column(Integer, ForeignKey(BaseDatasetORM.id, ondelete="cascade"), primary_key=True)

    __mapper_args__ = {
        "polymorphic_identity": "reaction",
    }


dataset_records_select = [
    select(
        BaseDatasetORM.id.label("dataset_id"),
        ReactionDatasetRecordItemORM.entry_name.label("entry_name"),
        ReactionDatasetRecordItemORM.specification_name.label("specification_name"),
        ReactionDatasetRecordItemORM.record_id.label("record_id"),
    )
    .join(ReactionDatasetRecordItemORM, BaseDatasetORM.id == ReactionDatasetRecordItemORM.dataset_id)
    .where(BaseDatasetORM.dataset_type == "reaction")
]



"""Add projects views

Revision ID: 0390661b91bf
Revises: 56ef8ac1765b
Create Date: 2025-04-17 10:26:09.369279

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0390661b91bf"
down_revision = "56ef8ac1765b"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        sa.text(
            """
                create or replace view project_records_view as  SELECT project_record.project_id,
                    project_record.record_id
                   FROM project_record
                UNION
                 SELECT project_record.project_id,
                    record_children_view.child_id AS record_id
                   FROM (project_record
                     JOIN record_children_view ON ((project_record.record_id = record_children_view.parent_id)))
                UNION
                 SELECT project_dataset.project_id,
                    dataset_direct_records_view.record_id
                   FROM (project_dataset
                     JOIN dataset_direct_records_view ON ((project_dataset.dataset_id = dataset_direct_records_view.dataset_id))); 
    """
        )
    )


def downgrade():
    op.execute(sa.text("DROP VIEW IF EXISTS project_records_view"))
    pass



import time
import traceback
from typing import Dict, Any

from flask import g, request, current_app, jsonify, Response
from jwt.exceptions import InvalidSubjectError
from werkzeug.exceptions import InternalServerError, HTTPException

from qcfractal.flask_app import storage_socket
from qcportal.exceptions import (
    UserReportableError,
    AuthenticationFailure,
    ComputeManagerError,
    AuthorizationFailure,
    SecurityNotEnabledError,
)
from .home_v1 import home_v1


#####################################################################
# This registers "global" error handlers and before/after
# request functions. Not that we are using _app_ decorators,
# which make them global, even though we are using a decorator
# from a specific blueprint
#####################################################################


@home_v1.before_app_request
def before_request_func():
    # Store timing information in the request/app context
    # g here refers to flask.g
    g.request_start = time.time()

    if request.data:
        g.request_bytes = len(request.data)
    else:
        g.request_bytes = 0


@home_v1.after_app_request
def after_request_func(response: Response):
    #################################################################
    # NOTE: Do not touch response.response! It may mess up streaming
    #       responses and result in no content being sent
    #################################################################

    # Determine the time the request took
    # g here refers to flask.g

    request_duration = time.time() - g.request_start

    log_access = current_app.config["QCFRACTAL_CONFIG"].log_access
    if log_access:
        # What we are going to log to the DB
        log: Dict[str, Any] = {}

        log["module"] = request.blueprint
        log["method"] = request.method

        # Replace null in URI (since a malevolent user can do that)
        log["full_uri"] = request.path.replace("\0", "\\0")

        # get the real IP address behind a proxy or ngnix
        real_ip = request.headers.get("X-Real-IP", None)

        # The IP address is the last address listed in access_route, which
        # comes from the X-FORWARDED-FOR header
        # (If access_route is empty, use the original request ip)
        if real_ip is None:
            real_ip = request.access_route[-1] if len(request.access_route) > 0 else request.remote_addr

        if real_ip:
            log["ip_address"] = real_ip

        log["user_agent"] = request.headers.get("User-Agent", "")

        log["request_bytes"] = 0 if g.request_bytes is None else g.request_bytes
        log["request_duration"] = request_duration
        log["user_id"] = g.get("user_id", None)

        response_bytes = response.content_length
        log["response_bytes"] = 0 if response_bytes is None else response_bytes

        storage_socket.serverinfo.save_access(log)
        current_app.logger.debug(
            f"{request.method} {request.blueprint}: {g.request_bytes} -> {response_bytes} [{request_duration*1000:.1f}ms]"
        )

    return response


@home_v1.app_errorhandler(InternalServerError)
def handle_internal_error(error):
    # For otherwise unhandled errors
    # Do not report the details to the user. Instead, log it,
    # and send the user the error id

    # Copy the headers to a dict, and remove the JWT stuff
    headers = dict(request.headers.items())
    headers.pop("Authorization", None)

    tb = traceback.format_exc()

    error_log = {
        "error_text": tb,
        "user_id": g.get("user_id", None),
        "request_path": request.full_path,
        "request_headers": str(headers),
        "request_body": str(request.data)[:8192],
    }

    # Log it to the internal error table
    err_id = storage_socket.serverinfo.save_error(error_log)

    # Should we hide the error from the user?
    hide = current_app.config["QCFRACTAL_CONFIG"].hide_internal_errors

    if hide:
        msg = error.description + f"  **Refer to internal error id {err_id} when asking your admin**"
        return jsonify(msg=msg), error.code
    else:
        return jsonify(msg=tb), error.code


@home_v1.app_errorhandler(HTTPException)
def handle_http_exception(error):
    # This handles many errors, such as NotFound, Unauthorized, etc
    # These are all reportable to the user
    return jsonify(msg=str(error)), error.code


@home_v1.app_errorhandler(UserReportableError)
def handle_userreport_error(error):
    # This handles any errors that are reportable to the user
    return jsonify(msg=str(error)), 400


@home_v1.app_errorhandler(AuthenticationFailure)
def handle_authentication_error(error):
    # This handles Authentication errors (invalid user, password, etc)
    # Or if the user tries to request a resource without being logged in
    return (
        jsonify(
            msg=str(error),
            user_id=g.user_id if "user_id" in g else None,
            username=g.username if "username" in g else None,
        ),
        401,
    )


@home_v1.app_errorhandler(AuthorizationFailure)
def handle_authorization_error(error: AuthorizationFailure):
    # This handles when a logged-in user does not have access to something
    return (
        jsonify(
            msg=f"Forbidden: {str(error)}",
            user_id=g.user_id if "user_id" in g else None,
            username=g.username if "username" in g else None,
        ),
        403,
    )


@home_v1.app_errorhandler(SecurityNotEnabledError)
def handle_compute_manager_error(error: SecurityNotEnabledError):
    # Handle compute manager errors
    return jsonify(msg=str(error)), 401


@home_v1.app_errorhandler(ComputeManagerError)
def handle_compute_manager_error(error: ComputeManagerError):
    # Handle compute manager errors
    return jsonify(msg=str(error)), 400


@home_v1.app_errorhandler(InvalidSubjectError)
def handle_old_tokens(error):
    # Handle old tokens that have integers as the subject
    # Just say they have been expired, and you need to login again
    return jsonify(msg="Token has expired"), 401



from __future__ import annotations

import hashlib
import io
import logging
import os
import re
import tarfile
from collections import defaultdict
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import requests
from sqlalchemy import and_, or_, func, select, delete
from sqlalchemy.orm import load_only

import qcfractal
from qcfractal.components.auth.db_models import UserIDMapSubquery
from qcfractal.db_socket.helpers import get_query_proj_options
from qcportal.serverinfo import (
    AccessLogQueryFilters,
    AccessLogSummaryFilters,
    ErrorLogQueryFilters,
)
from qcportal.utils import now_at_utc
from .db_models import AccessLogORM, InternalErrorLogORM, MessageOfTheDayORM, ServerStatsMetadataORM

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Dict, Any, List, Optional

# GeoIP2 package is optional
try:
    import geoip2.database

    geoip2_found = True
except ImportError:
    geoip2_found = False


class ServerInfoSocket:
    """
    Socket for managing/querying server logs and information
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)

        self._geoip2_dir = root_socket.qcf_config.geoip2_dir
        self._geoip2_file_path = os.path.join(self._geoip2_dir, root_socket.qcf_config.geoip2_filename)
        self._maxmind_license_key = root_socket.qcf_config.maxmind_license_key

        self._geolocate_accesses_frequency = 120  # two minutes should be ok?
        self._update_geoip2_frequency = 60 * 60 * 24  # one day

        # Set up access logging
        self._access_log_enabled = root_socket.qcf_config.log_access
        self._delete_access_log_frequency = 60 * 60 * 24  # one day
        self._access_log_keep = root_socket.qcf_config.access_log_keep
        self._geoip2_enabled = geoip2_found and self._access_log_enabled

        # MOTD contents
        self._load_motd()

        if self._access_log_enabled:
            os.makedirs(self._geoip2_dir, exist_ok=True)
            if not geoip2_found:
                self._logger.info(
                    "GeoIP2 package not found. To include locations in access logs, install the geoip2 package"
                )

        ################################
        # Add various internal jobs
        ################################
        with self.root_socket.session_scope() as session:
            # updating geoip2 file
            # Only add this if we have the maxmind license key
            if self._geoip2_enabled and self._maxmind_license_key:
                with self.root_socket.optional_session(session) as session:
                    self.root_socket.internal_jobs.add(
                        "update_geoip2_file",
                        now_at_utc(),
                        "serverinfo.update_geoip2_file",
                        {},
                        user_id=None,
                        unique_name=True,
                        repeat_delay=self._update_geoip2_frequency,
                        session=session,
                    )

            # Doing processing of access log
            if self._geoip2_enabled:
                self.root_socket.internal_jobs.add(
                    "geolocate_accesses",
                    now_at_utc(),
                    "serverinfo.geolocate_accesses",
                    {},
                    user_id=None,
                    unique_name=True,
                    repeat_delay=self._geolocate_accesses_frequency,
                    session=session,
                )

            # Deleting old access logs
            if self._access_log_keep > 0 and self._access_log_enabled:
                self.root_socket.internal_jobs.add(
                    "delete_old_access_log",
                    now_at_utc() + timedelta(seconds=2.0),
                    "serverinfo.delete_old_access_logs",
                    {},
                    user_id=None,
                    unique_name=True,
                    repeat_delay=self._delete_access_log_frequency,
                    session=session,
                )

    def update_geoip2_file(self) -> None:
        # Possible to reach this if we changed the settings, but have a job still in the queue
        if not (self._geoip2_enabled and self._maxmind_license_key):
            return

        # Session is not needed, but must be consistent with other jobs
        maxmind_license_key = self._maxmind_license_key

        if not maxmind_license_key:
            self._logger.warning("No maxmind license key provided. Cannot update geoip2 database")
            return

        base_url = f"https://download.maxmind.com/app/geoip_download?edition_id=GeoLite2-City&license_key={maxmind_license_key}"
        db_url = base_url + "&suffix=tar.gz"  # Actual DB file
        sha256_url = base_url + "&suffix=tar.gz.sha256"  # File with the sha256 hash

        self._logger.info("Checking if update of geoip2 database is needed")

        # According to the docs, we can use a head request and look at the last-modified header
        r = requests.head(db_url)
        remote_last_modified = r.headers.get("content-disposition")
        m = re.match(r".*GeoLite2-City_(\d{8}).tar.gz.*", remote_last_modified)

        if not m:
            raise RuntimeError(
                "Could not get sane filename from maxmind server. Account issue? Content-disposition: "
                + remote_last_modified
            )

        remote_last_modified = m.group(1)

        date_path = os.path.join(self._geoip2_dir, "last_modified.txt")

        # What version do we have already?
        if os.path.exists(date_path):
            with open(date_path, "r") as f:
                local_last_modified = f.read()
        else:
            local_last_modified = "00000000"

        self._logger.debug(f"Maxmind GeoIP2 last modified: {remote_last_modified}")
        self._logger.debug(f"Local GeoIP2 version: {local_last_modified}")

        if remote_last_modified <= local_last_modified:
            self._logger.info("Update of GeoIP2 database not needed")
            return

        else:
            self._logger.info("Update of GeoIP2 database required")

        geoip_file_data = requests.get(db_url)
        geoip_file_sha256 = requests.get(sha256_url).text

        local_hash = hashlib.sha256(geoip_file_data.content).hexdigest()

        # Is like a file: [hash] [filename]
        # We only want the hash
        expected_hash = geoip_file_sha256.split()[0]
        if local_hash != expected_hash:
            self._logger.warning(f"Hashes for geoip2 data do not match. Expected {expected_hash}, got {local_hash}")

        fileobj = io.BytesIO(geoip_file_data.content)

        # Extract contents, but remove the subdirectory
        with tarfile.open(fileobj=fileobj, mode="r:gz") as tar:
            db_date = None
            to_extract = []
            for ti in tar.getmembers():
                if ti.isfile():
                    ti.name = os.path.basename(ti.name)  # Remove the directory
                    to_extract.append(ti)
                if ti.isdir():
                    db_date = ti.name.split("_")[1]

            tar.extractall(path=self._geoip2_dir, members=to_extract)

            # Write the date of the subdir into the last_modified file
            with open(os.path.join(self._geoip2_dir, "last_modified.txt"), "w") as f:
                f.write(db_date)

        self._logger.info(f"Geoip database (date {db_date}) downloaded and extracted to {self._geoip2_dir}")

    def geolocate_accesses(self, session: Session) -> None:
        """
        Finds and updates accesses which haven't been processed for geolocation data
        """

        # Possible to reach this if we changed the settings, but have a job still in the queue
        if not self._geoip2_enabled:
            return

        if not os.path.exists(self._geoip2_file_path):
            self._logger.warning(
                "GeoIP2 database file not found. Cannot add location data to accesses. "
                "May need to wait for the updater job to run"
            )
            return

        stmt = select(ServerStatsMetadataORM).where(ServerStatsMetadataORM.name == "last_geolocated_date")
        last_geolocated_date = session.execute(stmt).scalar_one_or_none()

        access_stmt = select(AccessLogORM)
        access_stmt = access_stmt.options(load_only(AccessLogORM.timestamp, AccessLogORM.ip_address))
        access_stmt = access_stmt.where(AccessLogORM.ip_address.is_not(None))

        if last_geolocated_date:
            access_stmt = access_stmt.where(AccessLogORM.timestamp > last_geolocated_date.date_value)

        access_stmt = access_stmt.order_by(AccessLogORM.timestamp.asc())

        to_process = session.execute(access_stmt).scalars().all()
        self._logger.info(f"Found {len(to_process)} accesses to process")

        if not to_process:
            return

        distinct_ip = set(x.ip_address for x in to_process)
        self._logger.info(f"Found {len(distinct_ip)} distinct ip addresses to process")

        geo_reader = geoip2.database.Reader(self._geoip2_file_path)

        def _lookup_ip(ip_address: str) -> Dict[str, Any]:
            out: Dict[str, Any] = {}
            try:
                loc_data = geo_reader.city(ip_address)
                out["country_code"] = loc_data.country.iso_code
                out["subdivision"] = loc_data.subdivisions.most_specific.name
                out["city"] = loc_data.city.name
                out["ip_lat"] = loc_data.location.latitude
                out["ip_long"] = loc_data.location.longitude
            except:
                pass

            return out

        geo_data = {ip: _lookup_ip(ip) for ip in distinct_ip}

        for access in to_process:
            geo = geo_data.get(access.ip_address, {})
            access.country_code = geo.get("country_code")
            access.subdivision = geo.get("subdivision")
            access.city = geo.get("city")
            access.ip_lat = geo.get("ip_lat")
            access.ip_long = geo.get("ip_long")

        # Update the last geolocated date
        if last_geolocated_date is None:
            last_geolocated_date = ServerStatsMetadataORM(name="last_geolocated_date")
            session.add(last_geolocated_date)

        last_geolocated_date.date_value = to_process[-1].timestamp

        session.commit()

    def delete_old_access_logs(self, session: Session) -> None:
        """
        Deletes old access logs (as defined by the configuration)
        """

        # we check when adding the job, but double check here
        if self._access_log_keep <= 0 or not self._access_log_enabled:
            return

        before = now_at_utc() - timedelta(seconds=self._access_log_keep)
        num_deleted = self.delete_access_logs(before, session=session)
        self._logger.info(f"Deleted {num_deleted} access logs before {before}")

    def _load_motd(self, *, session: Optional[Session] = None):
        stmt = select(MessageOfTheDayORM).order_by(MessageOfTheDayORM.id)
        with self.root_socket.optional_session(session, True) as session:
            motd_orm = session.execute(stmt).scalar_one_or_none()
            if motd_orm is None:
                self._motd = ""
            else:
                self._motd = motd_orm.motd

        self._motd_time = now_at_utc()

    def set_motd(self, new_motd: str, *, session: Optional[Session] = None):
        stmt = select(MessageOfTheDayORM).order_by(MessageOfTheDayORM.id)
        with self.root_socket.optional_session(session) as session:
            motd_orm = session.execute(stmt).scalar_one_or_none()
            if motd_orm is None:
                motd_orm = MessageOfTheDayORM(motd=new_motd)
                session.add(motd_orm)
            else:
                motd_orm.motd = new_motd

        self._motd = new_motd
        self._motd_time = now_at_utc()

    def get_motd(self, *, session: Optional[Session] = None):
        # If file is updated, reload it
        # Only load every 10 seconds though
        now = now_at_utc()
        checktime = self._motd_time + timedelta(seconds=10)

        if now > checktime:
            self._load_motd(session=session)

        return self._motd

    def save_access(self, log_data: Dict[str, Any], *, session: Optional[Session] = None) -> None:
        """
        Saves information about a request/access to the database

        Parameters
        ----------
        log_data
            Dictionary of data to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.
        """

        with self.root_socket.optional_session(session) as session:
            log = AccessLogORM(**log_data)
            session.add(log)

    def save_error(self, error_data: Dict[str, Any], *, session: Optional[Session] = None) -> int:
        """
        Saves information about an internal error to the database

        Parameters
        ----------
        error_data
            Dictionary of error data to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            The id of the newly-created error
        """

        log = InternalErrorLogORM(**error_data, qcfractal_version=qcfractal.__version__)
        with self.root_socket.optional_session(session) as session:
            session.add(log)
            session.flush()
            return log.id

    def query_access_log(
        self,
        query_data: AccessLogQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        """
        General query of server access logs

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        The entries will be returned in chronological order, with the most
        recent being first.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of access log dictionaries
        """

        proj_options = get_query_proj_options(AccessLogORM, query_data.include, query_data.exclude)

        stmt = select(AccessLogORM)

        and_query = []
        if query_data.module:
            and_query.append(AccessLogORM.module.in_(query_data.module))
        if query_data.method:
            method = [x.upper() for x in query_data.method]
            and_query.append(AccessLogORM.method.in_(method))
        if query_data.before:
            and_query.append(AccessLogORM.timestamp <= query_data.before)
        if query_data.after:
            and_query.append(AccessLogORM.timestamp >= query_data.after)

        if query_data.user:
            stmt = stmt.join(UserIDMapSubquery)

            int_ids = {x for x in query_data.user if isinstance(x, int) or x.isdecimal()}
            str_names = set(query_data.user) - int_ids

            and_query.append(or_(UserIDMapSubquery.username.in_(str_names), UserIDMapSubquery.id.in_(int_ids)))

        with self.root_socket.optional_session(session, True) as session:
            stmt = stmt.where(and_(True, *and_query))
            stmt = stmt.options(*proj_options)

            if query_data.cursor is not None:
                stmt = stmt.where(AccessLogORM.id < query_data.cursor)

            stmt = stmt.order_by(AccessLogORM.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(AccessLogORM.id)
            results = session.execute(stmt).scalars().all()
            result_dicts = [x.model_dict() for x in sorted(results, key=lambda x: x.timestamp, reverse=True)]

        return result_dicts

    def query_access_summary(
        self,
        query_data: AccessLogSummaryFilters,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        General query of server access logs, returning aggregate data

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A dictionary containing summary data
        """

        and_query = []
        if query_data.before:
            and_query.append(AccessLogORM.timestamp <= query_data.before)
        if query_data.after:
            and_query.append(AccessLogORM.timestamp >= query_data.after)

        result_dict = defaultdict(list)
        with self.root_socket.optional_session(session, True) as session:
            if query_data.group_by == "user":
                group_col = UserIDMapSubquery.username.label("group_col")
            elif query_data.group_by == "day":
                group_col = func.to_char(AccessLogORM.timestamp.op("AT TIME ZONE")("UTC"), "YYYY-MM-DD").label(
                    "group_col"
                )
            elif query_data.group_by == "hour":
                group_col = func.to_char(AccessLogORM.timestamp.op("AT TIME ZONE")("UTC"), "YYYY-MM-DD HH24").label(
                    "group_col"
                )
            elif query_data.group_by == "country":
                group_col = AccessLogORM.country_code.label("group_col")
            elif query_data.group_by == "subdivision":
                group_col = AccessLogORM.subdivision.label("group_col")
            else:
                raise RuntimeError(f"Unknown group_by: {query_data.group_by}")

            stmt = select(
                group_col,
                AccessLogORM.module,
                AccessLogORM.method,
                func.count(AccessLogORM.id),
                func.min(AccessLogORM.request_duration),
                func.percentile_disc(0.25).within_group(AccessLogORM.request_duration),
                func.percentile_disc(0.5).within_group(AccessLogORM.request_duration),
                func.percentile_disc(0.75).within_group(AccessLogORM.request_duration),
                func.percentile_disc(0.95).within_group(AccessLogORM.request_duration),
                func.max(AccessLogORM.request_duration),
                func.min(AccessLogORM.response_bytes),
                func.percentile_disc(0.25).within_group(AccessLogORM.response_bytes),
                func.percentile_disc(0.5).within_group(AccessLogORM.response_bytes),
                func.percentile_disc(0.75).within_group(AccessLogORM.response_bytes),
                func.percentile_disc(0.95).within_group(AccessLogORM.response_bytes),
                func.max(AccessLogORM.response_bytes),
            )

            stmt = stmt.where(and_(True, *and_query)).group_by(AccessLogORM.module, AccessLogORM.method, "group_col")

            if query_data.group_by == "user":
                stmt = stmt.join(UserIDMapSubquery)

            results = session.execute(stmt).all()

            # What comes out is a tuple in order of the specified columns
            # We group into a dictionary where the key is the date, and the value
            # is a dictionary with the rest of the information
            for row in results:
                d = {
                    "module": row[1],
                    "method": row[2],
                    "count": row[3],
                    "request_duration_info": row[4:10],
                    "response_bytes_info": row[10:16],
                }
                result_dict[row[0]].append(d)

        # replace None with "_none_"
        if None in result_dict:
            if "_none_" in result_dict:
                raise RuntimeError("Key _none_ already exists. Weird username or country?")
            result_dict["_none_"] = result_dict.pop(None)

        return dict(result_dict)

    def query_error_log(
        self,
        query_data: ErrorLogQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        """
        General query of server internal error logs

        All search criteria are merged via 'and'. Therefore, records will only
        be found that match all the criteria.

        The entries will be returned in chronological order, with the most
        recent being first.

        Parameters
        ----------
        query_data
            Fields/filters to query for
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            A list of errors (as dictionaries) that were found in the database.
        """

        and_query = []
        stmt = select(InternalErrorLogORM)

        if query_data.error_id:
            and_query.append(InternalErrorLogORM.id.in_(query_data.error_id))
        if query_data.before:
            and_query.append(InternalErrorLogORM.error_date <= query_data.before)
        if query_data.after:
            and_query.append(InternalErrorLogORM.error_date >= query_data.after)
        if query_data.user:
            stmt = stmt.join(UserIDMapSubquery)

            int_ids = {x for x in query_data.user if isinstance(x, int) or x.isdecimal()}
            str_names = set(query_data.user) - int_ids

            and_query.append(or_(UserIDMapSubquery.username.in_(str_names), UserIDMapSubquery.id.in_(int_ids)))

        with self.root_socket.optional_session(session, True) as session:
            stmt = stmt.where(and_(True, *and_query))

            if query_data.cursor is not None:
                stmt = stmt.where(InternalErrorLogORM.id < query_data.cursor)

            stmt = stmt.order_by(InternalErrorLogORM.id.desc())
            stmt = stmt.limit(query_data.limit)
            stmt = stmt.distinct(InternalErrorLogORM.id)
            results = session.execute(stmt).scalars().all()
            result_dicts = [x.model_dict() for x in sorted(results, key=lambda x: x.error_date, reverse=True)]

        return result_dicts

    def delete_access_logs(self, before: datetime, *, session: Optional[Session] = None) -> int:
        """
        Deletes access logs that were created before a certain date & time

        Parameters
        ----------
        before
            Delete access logs before this time
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
            The number of deleted entries
        """

        with self.root_socket.optional_session(session, False) as session:
            stmt = delete(AccessLogORM).where(AccessLogORM.timestamp < before)
            r = session.execute(stmt)
            return r.rowcount

    def delete_error_logs(self, before: datetime, *, session: Optional[Session] = None) -> int:
        """
        Deletes error entries that were created before a certain date & time

        Parameters
        ----------
        before
            Delete error entries before this time
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
            The number of deleted entries
        """

        with self.root_socket.optional_session(session, False) as session:
            stmt = delete(InternalErrorLogORM).where(InternalErrorLogORM.error_date < before)
            r = session.execute(stmt)
            return r.rowcount



from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING

from sqlalchemy import select, literal, insert

from qcfractal.components.gridoptimization.record_db_models import GridoptimizationRecordORM
from qcportal.gridoptimization import GridoptimizationDatasetNewEntry, GridoptimizationSpecification
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.record_models import PriorityEnum
from .dataset_db_models import (
    GridoptimizationDatasetORM,
    GridoptimizationDatasetSpecificationORM,
    GridoptimizationDatasetEntryORM,
    GridoptimizationDatasetRecordItemORM,
)
from ..base_dataset_socket import BaseDatasetSocket

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Sequence, Iterable, Tuple


class GridoptimizationDatasetSocket(BaseDatasetSocket):
    # Used by the base class
    dataset_orm = GridoptimizationDatasetORM
    specification_orm = GridoptimizationDatasetSpecificationORM
    entry_orm = GridoptimizationDatasetEntryORM
    record_item_orm = GridoptimizationDatasetRecordItemORM
    record_orm = GridoptimizationRecordORM

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseDatasetSocket.__init__(
            self,
            root_socket,
        )

        self._logger = logging.getLogger(__name__)

    def _add_specification(
        self, session: Session, specification: GridoptimizationSpecification
    ) -> Tuple[InsertMetadata, Optional[int]]:
        return self.root_socket.records.gridoptimization.add_specification(specification, session=session)

    def _create_entries(
        self, session: Session, dataset_id: int, new_entries: Sequence[GridoptimizationDatasetNewEntry]
    ):
        all_entries = []
        for entry in new_entries:
            meta, mol_ids = self.root_socket.molecules.add_mixed([entry.initial_molecule], session=session)

            new_ent = GridoptimizationDatasetEntryORM(
                dataset_id=dataset_id,
                name=entry.name,
                comment=entry.comment,
                additional_keywords=entry.additional_keywords,
                additional_optimization_keywords=entry.additional_optimization_keywords,
                attributes=entry.attributes,
                initial_molecule_id=mol_ids[0],
            )

            all_entries.append(new_ent)

        return all_entries

    def _submit(
        self,
        session: Session,
        dataset_id: int,
        entry_orm: Iterable[GridoptimizationDatasetEntryORM],
        spec_orm: Iterable[GridoptimizationDatasetSpecificationORM],
        existing_records: Iterable[Tuple[str, str]],
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
    ) -> InsertCountsMetadata:

        n_inserted = 0
        n_existing = 0

        for spec in spec_orm:
            goopt_spec_obj = spec.specification.to_model(GridoptimizationSpecification)
            goopt_spec_input_dict = goopt_spec_obj.dict()

            for entry in entry_orm:
                if (entry.name, spec.name) in existing_records:
                    continue

                new_go_spec = copy.deepcopy(goopt_spec_input_dict)
                new_go_spec["keywords"].update(entry.additional_keywords)
                new_go_spec["optimization_specification"]["keywords"].update(entry.additional_optimization_keywords)

                go_spec = GridoptimizationSpecification(
                    optimization_specification=new_go_spec["optimization_specification"],
                    keywords=new_go_spec["keywords"],
                )

                meta, gridopt_ids = self.root_socket.records.gridoptimization.add(
                    initial_molecules=[entry.initial_molecule_id],
                    go_spec=go_spec,
                    compute_tag=compute_tag,
                    compute_priority=compute_priority,
                    creator_user=creator_user_id,
                    find_existing=find_existing,
                    session=session,
                )

                if meta.success:
                    rec = GridoptimizationDatasetRecordItemORM(
                        dataset_id=dataset_id,
                        entry_name=entry.name,
                        specification_name=spec.name,
                        record_id=gridopt_ids[0],
                    )
                    session.add(rec)

                n_inserted += meta.n_inserted
                n_existing += meta.n_existing

        return InsertCountsMetadata(n_inserted=n_inserted, n_existing=n_existing)

    def _copy_entries(
        self,
        session: Session,
        source_dataset_id: int,
        destination_dataset_id: int,
        entry_names: Optional[Iterable[str]] = None,
    ):
        select_stmt = select(
            literal(destination_dataset_id),
            self.entry_orm.name,
            self.entry_orm.comment,
            self.entry_orm.initial_molecule_id,
            self.entry_orm.additional_keywords,
            self.entry_orm.additional_optimization_keywords,
            self.entry_orm.attributes,
        )

        select_stmt = select_stmt.where(self.entry_orm.dataset_id == source_dataset_id)

        if entry_names is not None:
            select_stmt = select_stmt.where(self.entry_orm.name.in_(entry_names))

        stmt = insert(self.entry_orm)
        stmt = stmt.from_select(
            [
                self.entry_orm.dataset_id,
                self.entry_orm.name,
                self.entry_orm.comment,
                self.entry_orm.initial_molecule_id,
                self.entry_orm.additional_keywords,
                self.entry_orm.additional_optimization_keywords,
                self.entry_orm.attributes,
            ],
            select_stmt,
        )

        session.execute(stmt)



"""Merge projects and db views

Revision ID: 56ef8ac1765b
Revises: e98c86069950, 5ae71b845526
Create Date: 2025-04-16 09:27:07.198910

"""

# revision identifiers, used by Alembic.
revision = "56ef8ac1765b"
down_revision = ("e98c86069950", "5ae71b845526")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



import time

import pytest
import requests

from qcarchivetesting import test_users
from qcportal.auth import UserInfo


def get_qcf_cookie(cookies):
    for c in cookies:
        if c.name == "qcf_session":
            return c
    return None


def cookies_has_qcf_cookie(cookies):
    return get_qcf_cookie(cookies) is not None


@pytest.mark.parametrize("use_forms", [True, False])
def test_auth_session_login_logout(secure_snowflake, use_forms):
    username = "admin_user"
    password = test_users["admin_user"]["pw"]
    uri = secure_snowflake.get_uri()

    sess = requests.Session()  # will store cookies automatically

    # First, not logged in = unauthorized
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized

    # Now go through the browser login. This should set a cookie
    if use_forms:
        r = sess.post(f"{uri}/auth/v1/session_login", data={"username": username, "password": password})
    else:
        r = sess.post(f"{uri}/auth/v1/session_login", json={"username": username, "password": password})

    assert r.status_code == 200
    assert cookies_has_qcf_cookie(sess.cookies)

    # Can get to protected endpoint
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 200

    # Now logout - cookie is removed
    r = sess.post(f"{uri}/auth/v1/session_logout")
    assert r.status_code == 200
    assert not cookies_has_qcf_cookie(sess.cookies)

    # Not logged in anymore
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized


def test_auth_session_user_disabled(secure_snowflake):
    username = "submit_user"
    password = test_users["submit_user"]["pw"]
    uri = secure_snowflake.get_uri()

    sess = requests.Session()  # will store cookies automatically

    # Now go through the browser login. This should set a cookie
    r = sess.post(f"{uri}/auth/v1/session_login", json={"username": username, "password": password})

    assert r.status_code == 200
    assert cookies_has_qcf_cookie(sess.cookies)

    # Disable the user
    storage_socket = secure_snowflake.get_storage_socket()
    uinfo = storage_socket.users.get("submit_user")
    uinfo["enabled"] = False

    storage_socket.users.modify(UserInfo(**uinfo), as_admin=True)
    assert not get_qcf_cookie(sess.cookies).is_expired()

    # Not logged in anymore
    time.sleep(3.1)  # wait for server verification cache to expire
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized
    assert "is disabled" in r.text


def test_auth_session_expires(secure_snowflake):
    username = "admin_user"
    password = test_users["admin_user"]["pw"]
    uri = secure_snowflake.get_uri()

    sess = requests.Session()  # will store cookies automatically

    # First, not logged in = unauthorized
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized

    # Now go through the browser login. This should set a cookie
    r = sess.post(f"{uri}/auth/v1/session_login", json={"username": username, "password": password})

    assert r.status_code == 200
    assert cookies_has_qcf_cookie(sess.cookies)

    # Wait for expiration
    time.sleep(secure_snowflake._qcf_config.api.user_session_max_age + 1)
    assert get_qcf_cookie(sess.cookies).is_expired()

    # Not logged in anymore
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized

    sess.cookies.clear_expired_cookies()
    assert not cookies_has_qcf_cookie(sess.cookies)
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized


def test_auth_session_forged_expires(secure_snowflake):
    username = "admin_user"
    password = test_users["admin_user"]["pw"]
    uri = secure_snowflake.get_uri()

    sess = requests.Session()  # will store cookies automatically

    # First, not logged in = unauthorized
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized

    # Now go through the browser login. This should set a cookie
    r = sess.post(f"{uri}/auth/v1/session_login", json={"username": username, "password": password})

    assert r.status_code == 200
    assert cookies_has_qcf_cookie(sess.cookies)

    # Wait for expiration
    time.sleep(secure_snowflake._qcf_config.api.user_session_max_age + 1)

    assert get_qcf_cookie(sess.cookies).is_expired()

    # Make it not expired
    c = get_qcf_cookie(sess.cookies)
    c.expires = c.expires + 100000

    # Still not logged in
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized


@pytest.mark.slow
def test_auth_session_extension(secure_snowflake):
    username = "admin_user"
    password = test_users["admin_user"]["pw"]
    uri = secure_snowflake.get_uri()

    sess = requests.Session()  # will store cookies automatically

    # First, not logged in = unauthorized
    r = sess.get(f"{uri}/api/v1/information")
    assert r.status_code == 401  # unauthorized

    # Now go through the browser login. This should set a cookie
    r = sess.post(f"{uri}/auth/v1/session_login", json={"username": username, "password": password})

    assert r.status_code == 200
    assert cookies_has_qcf_cookie(sess.cookies)
    session_id = get_qcf_cookie(sess.cookies).value
    last_exp = get_qcf_cookie(sess.cookies).expires

    max_wait = 2 * int(secure_snowflake._qcf_config.api.user_session_max_age)
    for i in range(max_wait):
        time.sleep(2)
        r = sess.get(f"{uri}/api/v1/information")

        c = get_qcf_cookie(sess.cookies)

        # All session_ids should be the same
        assert not c.is_expired()
        assert c.value == session_id
        assert c.expires > last_exp
        last_exp = c.expires

        assert r.status_code == 200



from sqlalchemy import union, DDL, event, table, Column, Integer

from qcfractal.components.gridoptimization.record_db_models import (
    record_children_select as gridoptimization_children,
    record_direct_children_select as gridoptimization_direct_children,
)
from qcfractal.components.manybody.record_db_models import (
    record_children_select as manybody_children,
    record_direct_children_select as manybody_direct_children,
)
from qcfractal.components.neb.record_db_models import (
    record_children_select as neb_children,
    record_direct_children_select as neb_direct_children,
)
from qcfractal.components.optimization.record_db_models import (
    record_children_select as optimization_children,
    record_direct_children_select as optimization_direct_children,
)
from qcfractal.components.reaction.record_db_models import (
    record_children_select as reaction_children,
    record_direct_children_select as reaction_direct_children,
)
from qcfractal.components.torsiondrive.record_db_models import (
    record_children_select as torsiondrive_children,
    record_direct_children_select as torsiondrive_direct_children,
)
from qcfractal.db_socket.base_orm import BaseORM
from qcfractal.db_socket.db_views import view

RecordDirectChildrenView = view(
    "record_direct_children_view",
    BaseORM.metadata,
    union(
        *optimization_direct_children,
        *gridoptimization_direct_children,
        *torsiondrive_direct_children,
        *manybody_direct_children,
        *reaction_direct_children,
        *neb_direct_children,
    ),
)


RecordChildrenView = view(
    "record_children_view",
    BaseORM.metadata,
    union(
        *optimization_children,
        *gridoptimization_children,
        *torsiondrive_children,
        *manybody_children,
        *reaction_children,
        *neb_children,
    ),
)



from flask import g

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.base_models import ProjURLParameters
from qcportal.project_models import (
    ProjectAddBody,
    ProjectQueryModel,
    ProjectDeleteParams,
    ProjectDatasetAddBody,
    ProjectRecordAddBody,
    ProjectRecordImportBody,
    ProjectLinkDatasetBody,
    ProjectUnlinkDatasetsBody,
    ProjectLinkRecordBody,
    ProjectUnlinkRecordsBody,
)


@api_v1.route("/projects", methods=["GET"])
@wrap_global_route("projects", "read")
def list_project_v1():
    return storage_socket.projects.list()


@api_v1.route("/projects/<int:project_id>", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_v1(project_id: int, url_params: ProjURLParameters):
    return storage_socket.projects.get(project_id, url_params.include, url_params.exclude)


@api_v1.route("/projects/query", methods=["POST"])
@wrap_global_route("projects", "read")
def query_general_project_v1(body_data: ProjectQueryModel):
    with storage_socket.session_scope(True) as session:
        project_id = storage_socket.projects.lookup_id(body_data.project_name, session=session)
        return storage_socket.projects.get(project_id, body_data.include, body_data.exclude, session=session)


@api_v1.route("/projects/<int:project_id>", methods=["DELETE"])
@wrap_global_route("projects", "delete")
def delete_project_v1(project_id: int, url_params: ProjectDeleteParams):
    return storage_socket.projects.delete_project(
        project_id,
        delete_records=url_params.delete_records,
        delete_datasets=url_params.delete_datasets,
        delete_dataset_records=url_params.delete_dataset_records,
    )


########################
# Adding a project
########################
@api_v1.route("/projects", methods=["POST"])
@wrap_global_route("projects", "add")
def add_project_v1(body_data: ProjectAddBody):
    return storage_socket.projects.add(
        name=body_data.name,
        description=body_data.description,
        tagline=body_data.tagline,
        tags=body_data.tags,
        default_compute_tag=body_data.default_compute_tag,
        default_compute_priority=body_data.default_compute_priority,
        extras=body_data.extras,
        owner_user=g.username,
        existing_ok=body_data.existing_ok,
    )


#########################
# Getting info
#########################
@api_v1.route("/projects/<int:project_id>/status", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_status_v1(project_id: int):
    return storage_socket.projects.status(project_id)


@api_v1.route("/projects/<int:project_id>/record_metadata", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_record_metadata_v1(project_id: int):
    return storage_socket.projects.get_record_metadata(project_id)


@api_v1.route("/projects/<int:project_id>/dataset_metadata", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_dataset_metadata_v1(project_id: int):
    return storage_socket.projects.get_dataset_metadata(project_id)


#########################
# Datasets
#########################
@api_v1.route("/projects/<int:project_id>/datasets", methods=["POST"])
@wrap_global_route("projects", "modify")
def add_project_dataset_v1(project_id: int, body_data: ProjectDatasetAddBody):
    return storage_socket.projects.add_dataset(
        project_id=project_id,
        dataset_type=body_data.dataset_type,
        dataset_name=body_data.name,
        description=body_data.description,
        tagline=body_data.tagline,
        tags=body_data.tags,
        provenance=body_data.provenance,
        default_compute_tag=body_data.default_compute_tag,
        default_compute_priority=body_data.default_compute_priority,
        extras=body_data.extras,
        creator_user=g.username,
        existing_ok=body_data.existing_ok,
    )


@api_v1.route("/projects/<int:project_id>/datasets/link", methods=["POST"])
@wrap_global_route("projects", "modify")
def project_link_dataset_v1(project_id: int, body_data: ProjectLinkDatasetBody):
    return storage_socket.projects.link_dataset(
        project_id=project_id,
        dataset_id=body_data.dataset_id,
        name=body_data.name,
        description=body_data.description,
        tagline=body_data.tagline,
        tags=body_data.tags,
    )


@api_v1.route("/projects/<int:project_id>/datasets/unlink", methods=["POST"])
@wrap_global_route("projects", "modify")
def project_unlink_datasets_v1(project_id: int, body_data: ProjectUnlinkDatasetsBody):
    return storage_socket.projects.unlink_datasets(
        project_id=project_id,
        dataset_ids=body_data.dataset_ids,
        delete_datasets=body_data.delete_datasets,
        delete_dataset_records=body_data.delete_dataset_records,
    )


@api_v1.route("/projects/<int:project_id>/datasets/<int:dataset_id>", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_dataset_v1(project_id: int, dataset_id: int):
    return storage_socket.projects.get_dataset(project_id, dataset_id)


#########################
# Records
#########################
@api_v1.route("/projects/<int:project_id>/records", methods=["POST"])
@wrap_global_route("projects", "modify")
def add_project_record_v1(project_id: int, body_data: ProjectRecordAddBody):
    return storage_socket.projects.add_record(
        project_id,
        name=body_data.name,
        description=body_data.description,
        tags=body_data.tags,
        record_input=body_data.record_input,
        compute_tag=body_data.compute_tag,
        compute_priority=body_data.compute_priority,
        creator_user=g.username,
        find_existing=body_data.find_existing,
    )


@api_v1.route("/projects/<int:project_id>/records/import", methods=["POST"])
@wrap_global_route("projects", "modify")
def import_project_record_v1(project_id: int, body_data: ProjectRecordImportBody):
    return storage_socket.projects.import_record(
        project_id,
        name=body_data.name,
        description=body_data.description,
        tags=body_data.tags,
        record_data=body_data.record_data,
        creator_user=g.username,
    )


@api_v1.route("/projects/<int:project_id>/records/link", methods=["POST"])
@wrap_global_route("projects", "modify")
def project_link_record_v1(project_id: int, body_data: ProjectLinkRecordBody):
    return storage_socket.projects.link_record(
        project_id=project_id,
        record_id=body_data.record_id,
        name=body_data.name,
        description=body_data.description,
        tags=body_data.tags,
    )


@api_v1.route("/projects/<int:project_id>/records/unlink", methods=["POST"])
@wrap_global_route("projects", "modify")
def project_unlink_records_v1(project_id: int, body_data: ProjectUnlinkRecordsBody):
    return storage_socket.projects.unlink_records(
        project_id=project_id,
        record_ids=body_data.record_ids,
        delete_records=body_data.delete_records,
    )


@api_v1.route("/projects/<int:project_id>/records/<int:record_id>", methods=["GET"])
@wrap_global_route("projects", "read")
def get_project_record_v1(project_id: int, record_id: int, url_params: ProjURLParameters):
    return storage_socket.projects.get_record(
        project_id, record_id, include=url_params.include, exclude=url_params.exclude
    )


#########################
# Modifying metadata
#########################
# @api_v1.route("/projects/<string:project_type>/<int:project_id>", methods=["PATCH"])
# @wrap_global_route("projects", "modify")
# def modify_project_metadata_v1(project_type: str, project_id: int, body_data: DatasetModifyMetadata):
#    ds_socket = storage_socket.projects.get_socket(project_type)
#    return ds_socket.update_metadata(project_id, new_metadata=body_data)


#########################
# Attachments
#################################
@api_v1.route("/projects/<int:project_id>/attachments", methods=["GET"])
@wrap_global_route("projects", "read")
def fetch_project_attachments_v1(project_id: int):
    return storage_socket.projects.get_attachments(project_id)


@api_v1.route("/projects/<int:project_id>/attachments/<int:attachment_id>", methods=["DELETE"])
@wrap_global_route("projects", "modify")
def delete_project_attachment_v1(project_id: int, attachment_id: int):
    return storage_socket.projects.delete_attachment(project_id, attachment_id)



"""record extras not nullable

Revision ID: 03c96181c90f
Revises: 73b4838a6839
Create Date: 2024-10-30 11:07:37.641631

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "03c96181c90f"
down_revision = "73b4838a6839"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("""UPDATE base_record SET extras = '{}' WHERE extras IS NULL""")
    op.alter_column("base_record", "extras", existing_type=postgresql.JSONB(astext_type=sa.Text()), nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # No need to set extras to null - empty dict is still just fine

    op.alter_column("base_record", "extras", existing_type=postgresql.JSONB(astext_type=sa.Text()), nullable=True)
    # ### end Alembic commands ###



"""
Tests the general record socket
"""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

from qcelemental.models import FailedOperation, ComputeError

from qcfractal.components.optimization.testing_helpers import load_procedure_data as load_opt_procedure_data
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import load_procedure_data as load_sp_procedure_data
from qcfractal.testing_helpers import mname1
from qcfractalcompute.compress import compress_result
from qcportal.compression import decompress
from qcportal.record_models import PriorityEnum, RecordStatusEnum

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket


def populate_records_status(storage_socket: SQLAlchemySocket):
    """
    Populates the db with tasks in all statuses
    """

    manager_programs = {
        "qcengine": ["unknown"],
        "psi4": ["unknown"],
        "qchem": ["v3.0"],
        "rdkit": ["unknown"],
        "geometric": ["unknown"],
    }

    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=manager_programs,
        compute_tags=["tag1", "tag2", "tag3", "tag6"],
    )

    input_spec_0, molecule_0, result_data_0 = load_opt_procedure_data("opt_psi4_methane_sometraj")
    input_spec_1, molecule_1, result_data_1 = load_sp_procedure_data("sp_psi4_water_gradient")
    input_spec_2, molecule_2, result_data_2 = load_sp_procedure_data("sp_psi4_water_hessian")
    input_spec_3, molecule_3, result_data_3 = load_opt_procedure_data("opt_psi4_benzene")
    input_spec_4, molecule_4, result_data_4 = load_sp_procedure_data("sp_rdkit_water_energy")
    input_spec_5, molecule_5, result_data_5 = load_sp_procedure_data("sp_psi4_benzene_energy_2")
    input_spec_6, molecule_6, result_data_6 = load_sp_procedure_data("sp_psi4_water_energy")

    meta, id_0 = storage_socket.records.optimization.add(
        [molecule_0], input_spec_0, "tag0", PriorityEnum.normal, None, True
    )
    meta, id_1 = storage_socket.records.singlepoint.add(
        [molecule_1], input_spec_1, "tag1", PriorityEnum.high, None, True
    )
    meta, id_2 = storage_socket.records.singlepoint.add(
        [molecule_2], input_spec_2, "tag2", PriorityEnum.high, None, True
    )
    meta, id_3 = storage_socket.records.optimization.add(
        [molecule_3],
        input_spec_3,
        "tag3",
        PriorityEnum.high,
        None,
        True,
    )
    meta, id_4 = storage_socket.records.singlepoint.add(
        [molecule_4], input_spec_4, "tag4", PriorityEnum.normal, None, True
    )
    meta, id_5 = storage_socket.records.singlepoint.add(
        [molecule_5], input_spec_5, "tag5", PriorityEnum.normal, None, True
    )
    meta, id_6 = storage_socket.records.singlepoint.add(
        [molecule_6], input_spec_6, "tag6", PriorityEnum.normal, None, True
    )
    all_id = id_0 + id_1 + id_2 + id_3 + id_4 + id_5 + id_6

    # 0 = waiting   1 = complete   2 = running
    # 3 = error     4 = cancelled  5 = deleted
    # 6 = invalid

    # claim only the ones we want to be complete, running, or error (1, 2, 3, 6)
    # 6 needs to be complete to be invalidated
    tasks = storage_socket.tasks.claim_tasks(
        mname1.fullname, manager_programs, ["tag1", "tag2", "tag3", "tag6"], limit=4
    )
    assert len(tasks) == 4

    fop = FailedOperation(error=ComputeError(error_type="test_error", error_message="this is a test error"))

    # we don't send back the one we want to be 'running' still (#2)
    storage_socket.tasks.update_finished(
        mname1.fullname,
        {
            # tasks[1] is left running (corresponds to record 2)
            tasks[0]["id"]: compress_result(result_data_1.dict()),
            tasks[2]["id"]: compress_result(fop.dict()),
            tasks[3]["id"]: compress_result(result_data_6.dict()),
        },
    )

    # Add some more entries to the history of #3 (failing)
    for i in range(4):
        meta = storage_socket.records.reset(id_3)
        assert meta.success
        tasks = storage_socket.tasks.claim_tasks(
            mname1.fullname, manager_programs, ["tag1", "tag2", "tag3", "tag6"], limit=1
        )
        assert len(tasks) == 1
        assert tasks[0]["tag"] == "tag3"

        fop_compress = compress_result(fop.dict())
        storage_socket.tasks.update_finished(mname1.fullname, {tasks[0]["id"]: fop_compress})

    meta = storage_socket.records.cancel(id_4)
    assert meta.n_updated == 1
    meta = storage_socket.records.delete(id_5)
    assert meta.n_deleted == 1
    meta = storage_socket.records.invalidate(id_6)
    assert meta.n_updated == 1

    with storage_socket.session_scope() as session:
        all_rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert all_rec[0].status == RecordStatusEnum.waiting
        assert all_rec[1].status == RecordStatusEnum.complete
        assert all_rec[2].status == RecordStatusEnum.running
        assert all_rec[3].status == RecordStatusEnum.error
        assert all_rec[4].status == RecordStatusEnum.cancelled
        assert all_rec[5].status == RecordStatusEnum.deleted
        assert all_rec[6].status == RecordStatusEnum.invalid

    return all_id


def convert_to_plain_qcschema_result(result):
    """
    Converts a manager-mangled qcschema result into a plain qcschema result.

    Managers typically compress outputs and native files then store them in extras.
    This removes those and puts them back in the proper place
    """

    update = {}

    extras = deepcopy(result.extras)

    compressed_outputs = extras.pop("_qcfractal_compressed_outputs", None)
    compressed_native_files = extras.pop("_qcfractal_compressed_native_files", None)

    if compressed_outputs or compressed_native_files:
        update["extras"] = extras

    if compressed_outputs:
        # Keys are stdout, stderr, error, which match the fields of the result (I hope)
        for k, v in compressed_outputs.items():
            update[k] = decompress(v["data"], v["compression_type"])

    if compressed_native_files:
        update["native_files"] = {}
        for k, v in compressed_native_files.items():
            update["native_files"][k] = decompress(v["data"], v["compression_type"])

    if update:
        return result.copy(update=update)
    else:
        return result
